from __future__ import absolute_import
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:

import warnings

import numpy as np
import nibabel as nib

from nose.tools import assert_equal
from numpy.testing import (assert_array_almost_equal,
                           assert_array_equal,
                           assert_raises)

from .... import load_image
from ....core.image.image_spaces import (make_xyz_image, xyz_affine)
from ....fixes.nibabel import io_orientation
from ....io.nibcompat import get_header
from ....testing import funcfile
from ...slicetiming.timefuncs import st_43210, st_02413, st_42031
from ..affine import Rigid
from ..groupwise_registration import (Image4d, resample4d, FmriRealign4d,
                                      SpaceTimeRealign, SpaceRealign, Realign4d,
                                      Realign4dAlgorithm, make_grid)

IM = load_image(funcfile)
IMS = [nib.Nifti1Image(np.zeros((2, 3, 4, 5)), np.eye(4)) for i in range(4)]
for ix, imx in enumerate(IMS):
    get_header(imx)['pixdim'][4] = ix

def test_futurewarning():
    with warnings.catch_warnings(record=True) as warns:
        warnings.simplefilter('always')
        FmriRealign4d([IM], tr=2., slice_order='ascending')
        assert_equal(warns.pop(0).category, FutureWarning)


def test_scanner_time():
    im4d = Image4d(IM.get_data(), IM.affine, tr=3.,
                   slice_times=(0, 1, 2))
    assert_equal(im4d.scanner_time(0, 0), 0.)
    assert_equal(im4d.scanner_time(0, im4d.tr), 1.)


def test_slice_info():
    im4d = Image4d(IM.get_data(), IM.affine, tr=3.,
                   slice_times=(0, 1, 2), slice_info=(2, -1))
    assert_equal(im4d.slice_axis, 2)
    assert_equal(im4d.slice_direction, -1)


def test_slice_timing():
    affine = np.eye(4)
    affine[0:3, 0:3] = IM.affine[0:3, 0:3]
    im4d = Image4d(IM.get_data(), affine, tr=2., slice_times=0.0)
    x = resample4d(im4d, [Rigid() for i in range(IM.shape[3])])
    assert_array_almost_equal(im4d.get_data(), x)


def test_realign4d_no_time_interp():
    runs = [IM, IM]
    R = FmriRealign4d(runs, time_interp=False)
    assert R.slice_times == 0


def test_realign4d_ascending():
    runs = [IM, IM]
    R = FmriRealign4d(runs, tr=3, slice_order='ascending')
    assert_array_equal(R.slice_times, (0, 1, 2))
    assert R.tr == 3


def test_realign4d_descending():
    runs = [IM, IM]
    R = FmriRealign4d(runs, tr=3, slice_order='descending')
    assert_array_equal(R.slice_times, (2, 1, 0))
    assert R.tr == 3


def test_realign4d_ascending_interleaved():
    runs = [IM, IM]
    R = FmriRealign4d(runs, tr=3, slice_order='ascending', interleaved=True)
    assert_array_equal(R.slice_times, (0, 2, 1))
    assert R.tr == 3


def test_realign4d_descending_interleaved():
    runs = [IM, IM]
    R = FmriRealign4d(runs, tr=3, slice_order='descending', interleaved=True)
    assert_array_equal(R.slice_times, (1, 2, 0))
    assert R.tr == 3


def wrong_call(slice_times=None, slice_order=None, tr_slices=None,
               interleaved=None, time_interp=None):
    runs = [IM, IM]
    return FmriRealign4d(runs, tr=3, slice_times=slice_times,
                         slice_order=slice_order,
                         tr_slices=tr_slices,
                         interleaved=interleaved,
                         time_interp=time_interp)


def test_realign4d_incompatible_args():
    assert_raises(ValueError, wrong_call, slice_order=(0, 1, 2),
                  interleaved=False)
    assert_raises(ValueError, wrong_call, slice_times=(0, 1, 2),
                  slice_order='ascending')
    assert_raises(ValueError, wrong_call, slice_times=(0, 1, 2),
                  slice_order=(0, 1, 2))
    assert_raises(ValueError, wrong_call, slice_times=(0, 1, 2),
                  time_interp=True)
    assert_raises(ValueError, wrong_call, slice_times=(0, 1, 2),
                  time_interp=False)
    assert_raises(ValueError, wrong_call, time_interp=True)
    assert_raises(ValueError, wrong_call, slice_times=(0, 1, 2),
                  tr_slices=1)


def test_realign4d():
    """
    This tests whether realign4d yields the same results depending on
    whether the slice order is input explicitely or as
    slice_times='ascending'.

    Due to the very small size of the image used for testing (only 3
    slices), optimization is numerically unstable. It seems to make
    the default optimizer, namely scipy.fmin.fmin_ncg, adopt a random
    behavior. To work around the resulting inconsistency in results,
    we use nipy.optimize.fmin_steepest as the optimizer, although it's
    generally not recommended in practice.
    """
    runs = [IM, IM]
    orient = io_orientation(IM.affine)
    slice_axis = int(np.where(orient[:, 0] == 2)[0])
    R1 = SpaceTimeRealign(runs, tr=2., slice_times='ascending',
                          slice_info=slice_axis)
    R1.estimate(refscan=None, loops=1, between_loops=1, optimizer='steepest')
    nslices = IM.shape[slice_axis]
    slice_times = (2. / float(nslices)) * np.arange(nslices)
    R2 = SpaceTimeRealign(runs, tr=2., slice_times=slice_times,
                          slice_info=slice_axis)
    R2.estimate(refscan=None, loops=1, between_loops=1, optimizer='steepest')
    for r in range(2):
        for i in range(IM.shape[3]):
            assert_array_almost_equal(R1._transforms[r][i].translation,
                                      R2._transforms[r][i].translation)
            assert_array_almost_equal(R1._transforms[r][i].rotation,
                                      R2._transforms[r][i].rotation)
    for i in range(IM.shape[3]):
            assert_array_almost_equal(R1._mean_transforms[r].translation,
                                      R2._mean_transforms[r].translation)
            assert_array_almost_equal(R1._mean_transforms[r].rotation,
                                      R2._mean_transforms[r].rotation)


def test_realign4d_runs_with_different_affines():
    aff = xyz_affine(IM)
    aff2 = aff.copy()
    aff2[0:3, 3] += 5
    im2 = make_xyz_image(IM.get_data(), aff2, 'scanner')
    runs = [IM, im2]
    R = SpaceTimeRealign(runs, tr=2., slice_times='ascending', slice_info=2)
    R.estimate(refscan=None, loops=1, between_loops=1, optimizer='steepest')
    cor_im, cor_im2 = R.resample()
    assert_array_equal(xyz_affine(cor_im2), aff)


def test_realign4d_params():
    # Some tests for input parameters to realign4d
    R = Realign4d(IM, 3, [0, 1, 2], None) # No slice_info - OK
    assert_equal(R.tr, 3)
    # TR cannot be None
    assert_raises(ValueError, Realign4d, IMS[1], None, [0, 1, 2], None)
    # TR cannot be zero
    assert_raises(ValueError, Realign4d, IMS[1], 0, [0, 1, 2], None)
    # TR can be None if slice times are None
    R = Realign4d(IM, None, None)
    assert_equal(R.tr, 1)


def test_spacetimerealign_params():
    runs = [IM, IM]
    for slice_times in ('descending', '43210', st_43210, [2, 1, 0]):
        R = SpaceTimeRealign(runs, tr=3, slice_times=slice_times, slice_info=2)
        assert_array_equal(R.slice_times, (2, 1, 0))
        assert_equal(R.tr, 3)
    for slice_times in ('asc_alt_2', '02413', st_02413, [0, 2, 1]):
        R = SpaceTimeRealign(runs, tr=3, slice_times=slice_times, slice_info=2)
        assert_array_equal(R.slice_times, (0, 2, 1))
        assert_equal(R.tr, 3)
    for slice_times in ('desc_alt_2', '42031', st_42031, [1, 2, 0]):
        R = SpaceTimeRealign(runs, tr=3, slice_times=slice_times, slice_info=2)
        assert_array_equal(R.slice_times, (1, 2, 0))
        assert_equal(R.tr, 3)
    # Check changing axis
    R = SpaceTimeRealign(runs, tr=21, slice_times='ascending', slice_info=1)
    assert_array_equal(R.slice_times, np.arange(21))
    # Check slice_times and slice_info and TR required
    R = SpaceTimeRealign(runs, 3, 'ascending', 2) # OK
    assert_raises(ValueError, SpaceTimeRealign, runs, 3, None, 2)
    assert_raises(ValueError, SpaceTimeRealign, runs, 3, 'ascending', None)
    assert_raises(ValueError, SpaceTimeRealign, IMS[0], None, [0, 1, 2], 2)
    assert_raises(ValueError, SpaceTimeRealign, IMS[1], None, [0, 1, 2], 2)
    assert_raises(ValueError, SpaceTimeRealign, IMS[2:4], None, [0, 1, 2], 2)
    assert_raises(ValueError, SpaceTimeRealign, IMS[0], 'header-allow-1.0', [0, 1, 2], 2)
    R = SpaceTimeRealign(IMS[1], "header-allow-1.0", 'ascending', 2)
    assert_array_equal(R.tr, 1.0)
    # Test when TR and nslices are not the same
    R1 = SpaceTimeRealign(runs, tr=2., slice_times='ascending', slice_info=2)
    assert_array_equal(R1.slice_times, np.arange(3) / 3. * 2.)
    # Smoke test run
    R1.estimate(refscan=None, loops=1, between_loops=1, optimizer='steepest')


def reduced_dim(dim, subsampling, border):
    return max(1, int(np.ceil((dim - 2 * border) / float(subsampling))))


def test_lowlevel_params():
    runs = [IM, IM]
    R = SpaceTimeRealign(runs, tr=21, slice_times='ascending', slice_info=1)
    borders=(3,2,1)
    R.estimate(refscan=None, loops=1, between_loops=1, optimizer='steepest', borders=borders)
    # Test tighter borders for motion estimation
    r = Realign4dAlgorithm(R._runs[0], borders=borders)
    nvoxels = np.prod(np.array([reduced_dim(IM.shape[i], 1, borders[i]) for i in range(3)]))
    assert_array_equal(r.xyz.shape, (nvoxels, 3))
    # Test wrong argument types raise errors
    assert_raises(ValueError, Realign4dAlgorithm, R._runs[0], subsampling=(3,3,3,1))
    assert_raises(ValueError, Realign4dAlgorithm, R._runs[0], refscan='first')
    assert_raises(ValueError, Realign4dAlgorithm, R._runs[0], borders=(1,1,1,0))
    assert_raises(ValueError, Realign4dAlgorithm, R._runs[0], xtol=None)
    assert_raises(ValueError, Realign4dAlgorithm, R._runs[0], ftol='dunno')
    assert_raises(ValueError, Realign4dAlgorithm, R._runs[0], gtol=(.1,.1,.1))
    assert_raises(ValueError, Realign4dAlgorithm, R._runs[0], stepsize=None)
    assert_raises(ValueError, Realign4dAlgorithm, R._runs[0], maxiter=None)
    assert_raises(ValueError, Realign4dAlgorithm, R._runs[0], maxfun='none')


def _test_make_grid(dims, subsampling, borders, expected_nvoxels):
    x = make_grid(dims, subsampling, borders)
    assert_equal(x.shape[0], expected_nvoxels)


def test_make_grid_funfile():
    dims = IM.shape[0:3]
    borders = (3,2,1)
    nvoxels = np.prod(np.array([reduced_dim(dims[i], 1, borders[i]) for i in range(3)]))
    _test_make_grid(dims, (1,1,1), borders, nvoxels)


def test_make_grid_default():
    dims = np.random.randint(100, size=3) + 1
    _test_make_grid(dims, (1,1,1), (0,0,0), np.prod(dims))


def test_make_grid_random_subsampling():
    dims = np.random.randint(100, size=3) + 1
    subsampling = np.random.randint(5, size=3) + 1
    nvoxels = np.prod(np.array([reduced_dim(dims[i], subsampling[i], 0) for i in range(3)]))
    _test_make_grid(dims, subsampling, (0,0,0), nvoxels)


def test_make_grid_random_borders():
    dims = np.random.randint(100, size=3) + 1
    borders = np.minimum((dims - 1) / 2, np.random.randint(10, size=3))
    nvoxels = np.prod(np.array([reduced_dim(dims[i], 1, borders[i]) for i in range(3)]))
    _test_make_grid(dims, (1,1,1), borders, nvoxels)


def test_make_grid_full_monthy():
    dims = np.random.randint(100, size=3) + 1
    subsampling = np.random.randint(5, size=3) + 1
    borders = np.minimum((dims - 1) / 2, np.random.randint(10, size=3))
    nvoxels = np.prod(np.array([reduced_dim(dims[i], subsampling[i], borders[i]) for i in range(3)]))
    _test_make_grid(dims, subsampling, borders, nvoxels)


def test_spacerealign():
    # Check space-only realigner
    runs = [IM, IM]
    R = SpaceRealign(runs)
    assert_equal(R.tr, 1)
    assert_equal(R.slice_times, 0.)
    # Smoke test run
    R.estimate(refscan=None, loops=1, between_loops=1, optimizer='steepest')


def test_single_image():
    # Check we can use a single image as argument
    R = SpaceTimeRealign(IM, tr=3, slice_times='ascending', slice_info=2)
    R.estimate(refscan=None, loops=1, between_loops=1, optimizer='steepest')
    R = SpaceRealign(IM)
    R.estimate(refscan=None, loops=1, between_loops=1, optimizer='steepest')
    R = Realign4d(IM, 3, [0, 1, 2], (2, 1))
    R.estimate(refscan=None, loops=1, between_loops=1, optimizer='steepest')



""" Testing diagnostics.command module
"""
from __future__ import absolute_import

import os
from os.path import dirname, join as pjoin, isfile
import shutil

import numpy as np

import nibabel as nib
from nibabel import AnalyzeImage, Spm2AnalyzeImage, Nifti1Pair, Nifti1Image
from nibabel.tmpdirs import InTemporaryDirectory

from nipy import load_image
from nipy.io.nifti_ref import NiftiError
from ..commands import parse_fname_axes, tsdiffana, diagnose
from ..timediff import time_slice_diffs_image

from numpy.testing import (assert_almost_equal, assert_array_equal)

from nose import SkipTest
from nose.tools import (assert_true, assert_false, assert_raises,
                        assert_equal, assert_not_equal)

from nipy.testing import funcfile
from nipy.testing.decorators import needs_mpl_agg


def test_parse_fname_axes():
    # Test logic for setting time and slice axis defaults
    # We need real images for the tests because nipy will load them
    # For simplicity, we can create them
    shape = (4, 5, 6, 20)
    arr = np.arange(np.prod(shape), dtype=float).reshape(shape)
    zooms = (2., 3., 4., 2.1)
    with InTemporaryDirectory():
        for (img_class, ext) in ((AnalyzeImage, '.img'),
                                 (Spm2AnalyzeImage, '.img'),
                                 (Nifti1Pair, '.img'),
                                 (Nifti1Image, '.nii')):
            hdr = img_class.header_class()
            hdr.set_data_shape(shape)
            hdr.set_zooms(zooms)
            hdr.set_data_dtype(np.dtype(np.float64))
            nibabel_img = img_class(arr, None, hdr)
            # We so far haven't set any slice axis information
            for z_ext in ('', '.gz'):
                fname = 'image' + ext + z_ext
                nib.save(nibabel_img, fname)
                for in_time, in_sax, out_time, out_sax in (
                    (None, None, 't', 2),
                    (None, '0', 't', 0),
                    (None, 'i', 't', 'i'),
                    (None, '1', 't', 1),
                    (None, 'j', 't', 'j'),
                    ('k', 'j', 'k', 'j'),
                    ('k', None, 'k', 2)):
                    img, time_axis, slice_axis = parse_fname_axes(
                        fname,
                        in_time,
                        in_sax)
                    assert_equal(time_axis, out_time)
                    assert_equal(slice_axis, out_sax)
                    del img
            # For some images, we can set the slice dimension. This becomes the
            # default if input slice_axis is None
            if hasattr(hdr, 'set_dim_info'):
                for ax_no in range(3):
                    nibabel_img.get_header().set_dim_info(slice=ax_no)
                    nib.save(nibabel_img, fname)
                    img, time_axis, slice_axis = parse_fname_axes(fname,
                                                                  None,
                                                                  None)
                    assert_equal(time_axis, 't')
                    assert_equal(slice_axis, 'slice')
                    del img
            # Images other than 4D don't get the slice axis default
            for new_arr in (arr[..., 0], arr[..., None]):
                new_nib = img_class(new_arr, None, hdr)
                nib.save(new_nib, fname)
                assert_raises(ValueError, parse_fname_axes, fname, None, None)
                # But you can still set slice axis
                img, time_axis, slice_axis = parse_fname_axes(fname, None, 'j')
                assert_equal(time_axis, 't')
                assert_equal(slice_axis, 'j')
    # Non-analyze image types don't get the slice default
    nib_data = pjoin(dirname(nib.__file__), 'tests', 'data')
    mnc_4d_fname = pjoin(nib_data, 'minc1_4d.mnc')
    if isfile(mnc_4d_fname):
        assert_raises(ValueError, parse_fname_axes, mnc_4d_fname, None, None)
        # At the moment we can't even load these guys
        try:
            img, time_axis, slice_axis = parse_fname_axes(
                mnc_4d_fname, None, 'j')
        except ValueError: # failed load
            raise SkipTest('Hoping for a time when we can use MINC')
        # But you can still set slice axis (if we can load them)
        assert_equal(time_axis, 't')
        assert_equal(slice_axis, 'j')


class Args(object): pass


def check_axes(axes, img_shape, time_axis, slice_axis):
    # Check axes as expected for plot
    assert_equal(len(axes), 4)
    # First x axis is time point differences
    assert_array_equal(axes[0].xaxis.get_data_interval(),
                       [0, img_shape[time_axis]-2])
    # Last x axis is over slices
    assert_array_equal(axes[-1].xaxis.get_data_interval(),
                       [0, img_shape[slice_axis]-1])


@needs_mpl_agg
def test_tsdiffana():
    # Test tsdiffana command
    args = Args()
    img = load_image(funcfile)
    with InTemporaryDirectory() as tmpdir:
        args.filename = funcfile
        args.time_axis = None
        args.slice_axis = None
        args.write_results = False
        args.out_path = None
        args.out_fname_label = None
        args.out_file = 'test.png'
        check_axes(tsdiffana(args), img.shape, -1, -2)
        assert_true(isfile('test.png'))
        args.time_axis = 't'
        check_axes(tsdiffana(args), img.shape, -1, -2)
        args.time_axis = '3'
        check_axes(tsdiffana(args), img.shape, -1, -2)
        args.slice_axis = 'k'
        check_axes(tsdiffana(args), img.shape, -1, -2)
        args.slice_axis = '2'
        check_axes(tsdiffana(args), img.shape, -1, -2)
        args.time_axis = '0'
        check_axes(tsdiffana(args), img.shape, 0, -2)
        args.slice_axis = 't'
        check_axes(tsdiffana(args), img.shape, 0, -1)
        # Check absolute path works
        args.slice_axis = 'j'
        args.time_axis = 't'
        args.out_file = pjoin(tmpdir, 'test_again.png')
        check_axes(tsdiffana(args), img.shape, -1, -3)
        # Check that --out-images incompatible with --out-file
        args.write_results=True
        assert_raises(ValueError, tsdiffana, args)
        args.out_file=None
        # Copy the functional file to a temporary writeable directory
        os.mkdir('mydata')
        tmp_funcfile = pjoin(tmpdir, 'mydata', 'myfunc.nii.gz')
        shutil.copy(funcfile, tmp_funcfile)
        args.filename = tmp_funcfile
        # Check write-results generates expected images
        check_axes(tsdiffana(args), img.shape, -1, -3)
        assert_true(isfile(pjoin('mydata', 'tsdiff_myfunc.png')))
        max_img = load_image(pjoin('mydata', 'dv2_max_myfunc.nii.gz'))
        assert_equal(max_img.shape, img.shape[:-1])
        mean_img = load_image(pjoin('mydata', 'dv2_max_myfunc.nii.gz'))
        assert_equal(mean_img.shape, img.shape[:-1])
        exp_results = time_slice_diffs_image(img, 't', 'j')
        saved_results = np.load(pjoin('mydata', 'tsdiff_myfunc.npz'))
        for key in ('volume_means', 'slice_mean_diff2'):
            assert_array_equal(exp_results[key], saved_results[key])
        # That we can change out-path
        os.mkdir('myresults')
        args.out_path = 'myresults'
        check_axes(tsdiffana(args), img.shape, -1, -3)
        assert_true(isfile(pjoin('myresults', 'tsdiff_myfunc.png')))
        max_img = load_image(pjoin('myresults', 'dv2_max_myfunc.nii.gz'))
        assert_equal(max_img.shape, img.shape[:-1])
        # And out-fname-label
        args.out_fname_label = 'vr2'
        check_axes(tsdiffana(args), img.shape, -1, -3)
        assert_true(isfile(pjoin('myresults', 'tsdiff_vr2.png')))
        max_img = load_image(pjoin('myresults', 'dv2_max_vr2.nii.gz'))
        assert_equal(max_img.shape, img.shape[:-1])
        del max_img, mean_img


def check_diag_results(results, img_shape,
                       time_axis, slice_axis, ncomps,
                       out_path, froot, ext='.nii.gz'):

    S = img_shape[slice_axis]
    T = img_shape[time_axis]
    pca_shape = list(img_shape)
    pca_shape[time_axis] = ncomps
    assert_equal(results['pca'].shape, tuple(pca_shape))
    assert_equal(results['pca_res']['basis_projections'].shape,
                 tuple(pca_shape))
    # Roll pca axis last to test shape of output image
    ax_order = list(range(4))
    ax_order.remove(time_axis)
    ax_order.append(time_axis)
    rolled_shape = tuple(pca_shape[i] for i in ax_order)
    pca_img = load_image(pjoin(out_path, 'pca_' + froot + ext))
    assert_equal(pca_img.shape, rolled_shape)
    for prefix in ('mean', 'min', 'max', 'std'):
        fname = pjoin(out_path, prefix + '_' + froot + ext)
        img = load_image(fname)
        assert_equal(img.shape, rolled_shape[:-1])
    vars = np.load(pjoin(out_path, 'vectors_components_' + froot + '.npz'))
    assert_equal(set(vars),
                 set(['basis_vectors', 'pcnt_var', 'volume_means',
                      'slice_mean_diff2']))
    assert_equal(vars['volume_means'].shape, (T,))
    assert_equal(vars['basis_vectors'].shape, (T, T-1))
    assert_equal(vars['slice_mean_diff2'].shape, (T-1, S))


@needs_mpl_agg
def test_diagnose():
    args = Args()
    img = load_image(funcfile)
    with InTemporaryDirectory() as tmpdir:
        # Copy the functional file to a temporary writeable directory
        os.mkdir('mydata')
        tmp_funcfile = pjoin(tmpdir, 'mydata', 'myfunc.nii.gz')
        shutil.copy(funcfile, tmp_funcfile)
        args.filename = tmp_funcfile
        args.time_axis = None
        args.slice_axis = None
        args.out_path = None
        args.out_fname_label = None
        args.ncomponents = 10
        res = diagnose(args)
        check_diag_results(res, img.shape, 3, 2, 10, 'mydata', 'myfunc')
        args.slice_axis = 'j'
        res = diagnose(args)
        check_diag_results(res, img.shape, 3, 1, 10, 'mydata', 'myfunc')
        # Time axis is not going to work because we'd have to use up one of the
        # needed spatial axes
        args.time_axis = 'i'
        assert_raises(NiftiError, diagnose, args)
        args.time_axis = 't'
        # Check that output works
        os.mkdir('myresults')
        args.out_path = 'myresults'
        args.out_fname_label = 'myana'
        res = diagnose(args)
        check_diag_results(res, img.shape, 3, 1, 10, 'myresults', 'myana')



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
""" Test scripts

Run scripts and test output
"""
from __future__ import absolute_import

import os
from os.path import join as pjoin, isfile

import numpy as np

from nibabel.tmpdirs import InTemporaryDirectory

from nipy import load_image, save_image
from nipy.core.api import rollimg

from nose.tools import assert_true, assert_false, assert_equal, assert_raises

from ..testing import funcfile
from numpy.testing import decorators, assert_almost_equal

from nipy.testing.decorators import make_label_dec

from nibabel.optpkg import optional_package

matplotlib, HAVE_MPL, _ = optional_package('matplotlib')
needs_mpl = decorators.skipif(not HAVE_MPL, "Test needs matplotlib")
script_test = make_label_dec('script_test')

from .scriptrunner import ScriptRunner

runner = ScriptRunner(
    debug_print_var = 'NIPY_DEBUG_PRINT')
run_command = runner.run_command

@needs_mpl
@script_test
def test_nipy_diagnose():
    # Test nipy diagnose script
    fimg = load_image(funcfile)
    ncomps = 12
    with InTemporaryDirectory() as tmpdir:
        cmd = ['nipy_diagnose', funcfile,
               '--ncomponents={0}'.format(ncomps),
               '--out-path=' + tmpdir]
        run_command(cmd)
        for out_fname in ('components_functional.png',
                          'pcnt_var_functional.png',
                          'tsdiff_functional.png',
                          'vectors_components_functional.npz'):
            assert_true(isfile(out_fname))
        for out_img in ('max_functional.nii.gz',
                        'mean_functional.nii.gz',
                        'min_functional.nii.gz',
                        'std_functional.nii.gz'):
            img = load_image(out_img)
            assert_equal(img.shape, fimg.shape[:-1])
            del img
        pca_img = load_image('pca_functional.nii.gz')
        assert_equal(pca_img.shape, fimg.shape[:-1] + (ncomps,))
        vecs_comps = np.load('vectors_components_functional.npz')
        vec_diff = vecs_comps['slice_mean_diff2'].copy()# just in case
        assert_equal(vec_diff.shape, (fimg.shape[-1]-1, fimg.shape[2]))
        del pca_img, vecs_comps
    with InTemporaryDirectory() as tmpdir:
        # Check we can pass in slice and time flags
        s0_img = rollimg(fimg, 'k')
        save_image(s0_img, 'slice0.nii')
        cmd = ['nipy_diagnose', 'slice0.nii',
               '--ncomponents={0}'.format(ncomps),
               '--out-path=' + tmpdir,
               '--time-axis=t',
               '--slice-axis=0']
        run_command(cmd)
        pca_img = load_image('pca_slice0.nii')
        assert_equal(pca_img.shape, s0_img.shape[:-1] + (ncomps,))
        vecs_comps = np.load('vectors_components_slice0.npz')
        assert_almost_equal(vecs_comps['slice_mean_diff2'], vec_diff)
        del pca_img, vecs_comps


@needs_mpl
@script_test
def test_nipy_tsdiffana():
    # Test nipy_tsdiffana script
    out_png = 'ts_out.png'
    # Quotes in case of space in arguments
    with InTemporaryDirectory():
        for i, extras in enumerate(([],
                                    ['--time-axis=0'],
                                    ['--slice-axis=0'],
                                    ['--slice-axis=0', '--time-axis=1']
                                   )):
            out_png = 'ts_out{0}.png'.format(i)
            cmd = (['nipy_tsdiffana', funcfile] + extras +
                   ['--out-file=' + out_png])
            run_command(cmd)
            assert_true(isfile(out_png))
    # Out-file and write-results incompatible
    cmd = (['nipy_tsdiffana', funcfile, '--out-file=' + out_png,
            '--write-results'])
    assert_raises(RuntimeError,
                  run_command,
                  cmd)
    # Can save images
    cmd_root = ['nipy_tsdiffana', funcfile]
    with InTemporaryDirectory():
        os.mkdir('myresults')
        run_command(cmd_root + ['--out-path=myresults', '--write-results'])
        assert_true(isfile(pjoin('myresults', 'tsdiff_functional.png')))
        assert_true(isfile(pjoin('myresults', 'tsdiff_functional.npz')))
        assert_true(isfile(pjoin('myresults', 'dv2_max_functional.nii.gz')))
        assert_true(isfile(pjoin('myresults', 'dv2_mean_functional.nii.gz')))
        run_command(cmd_root + ['--out-path=myresults', '--write-results',
                                '--out-fname-label=vr2'])
        assert_true(isfile(pjoin('myresults', 'tsdiff_vr2.png')))
        assert_true(isfile(pjoin('myresults', 'tsdiff_vr2.npz')))
        assert_true(isfile(pjoin('myresults', 'dv2_max_vr2.nii.gz')))
        assert_true(isfile(pjoin('myresults', 'dv2_mean_vr2.nii.gz')))


@script_test
def test_nipy_3_4d():
    # Test nipy_3dto4d and nipy_4dto3d
    fimg = load_image(funcfile)
    N = fimg.shape[-1]
    out_4d = 'func4d.nii'
    with InTemporaryDirectory() as tmpdir:
        cmd = ['nipy_4dto3d', funcfile,  '--out-path=' + tmpdir]
        run_command(cmd)
        imgs_3d = ['functional_%04d.nii' % i for i in range(N)]
        for iname in imgs_3d:
            assert_true(isfile(iname))
        cmd = ['nipy_3dto4d'] + imgs_3d  + ['--out-4d=' + out_4d]
        run_command(cmd)
        fimg_back = load_image(out_4d)
        assert_almost_equal(fimg.get_data(), fimg_back.get_data())
        del fimg_back


@script_test
def test_nipy_4d_realign():
    # Test nipy_4d_realign script
    with InTemporaryDirectory():
        # Set matplotib agg backend
        with open("matplotlibrc", "wt") as fobj:
            fobj.write("backend : agg")
        cmd = ['nipy_4d_realign', '2.0', funcfile,
               '--slice_dim',  '2',  '--slice_dir', '-1', '--save_path', '.']
        run_command(cmd)



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
""" This module provides definitions of various hemodynamic response
functions (hrf).

In particular, it provides Gary Glover's canonical HRF, AFNI's default
HRF, and a spectral HRF.

The Glover HRF is based on:

@article{glover1999deconvolution,
  title={{Deconvolution of impulse response in event-related BOLD fMRI}},
  author={Glover, G.H.},
  journal={NeuroImage},
  volume={9},
  number={4},
  pages={416--429},
  year={1999},
  publisher={Orlando, FL: Academic Press, c1992-}
}

This parametrization is from fmristat:

http://www.math.mcgill.ca/keith/fmristat/

fmristat models the HRF as the difference of two gamma functions, ``g1``
and ``g2``, each defined by the timing of the gamma function peaks
(``pk1, pk2``) and the FWHMs (``width1, width2``):

   raw_hrf = g1(pk1, width1) - a2 * g2(pk2, width2)

where ``a2`` is the scale factor for the ``g2`` gamma function.  The
actual hrf is the raw hrf set to have an integral of 1. 

fmristat used ``pk1, width1, pk2, width2, a2 = (5.4 5.2 10.8 7.35
0.35)``.  These are parameters to match Glover's 1 second duration
auditory stimulus curves.  Glover wrote these as:

   y(t) = c1 * t**n1 * exp(t/t1) - a2 * c2 * t**n2 * exp(t/t2)

with ``n1, t1, n2, t2, a2 = (6.0, 0.9, 12, 0.9, 0.35)``,  and ``c1, c2`` being
``1/max(t**n1 * exp(t/t1)), 1/max(t**n2 * exp(t/t2)``.  The difference between
Glover's expression and ours is because we (and fmristat) use the peak location
and width to characterize the function rather than ``n1, t1``.  The values we
use are equivalent.  Specifically, in our formulation:

>>> n1, t1, c1 = gamma_params(5.4, 5.2)
>>> np.allclose((n1-1, t1), (6.0, 0.9), rtol=0.02)
True
>>> n2, t2, c2 = gamma_params(10.8, 7.35)
>>> np.allclose((n2-1, t2), (12.0, 0.9), rtol=0.02)
True
"""
from __future__ import division
from __future__ import absolute_import

from functools import partial

import numpy as np
import sympy

import scipy.stats as sps

from sympy.utilities.lambdify import implemented_function

from .utils import lambdify_t, T

def gamma_params(peak_location, peak_fwhm):
    """ Parameters for gamma density given peak and width

    TODO: where does the coef come from again.... check fmristat code

    From a peak location and peak FWHM, determine the parameters (shape,
    scale) of a Gamma density:

    f(x) = coef * x**(shape-1) * exp(-x/scale)

    The coefficient returned ensures that the f has integral 1 over
    [0,np.inf]

    Parameters
    ----------
    peak_location : float
       Location of the peak of the Gamma density
    peak_fwhm : float
       FWHM at the peak

    Returns
    -------
    shape : float
       Shape parameter in the Gamma density
    scale : float
       Scale parameter in the Gamma density
    coef : float
       Coefficient needed to ensure the density has integral 1.
    """
    shape_m1 = np.power(peak_location / peak_fwhm, 2) * 8 * np.log(2.0)
    scale = np.power(peak_fwhm, 2) / peak_location / 8 / np.log(2.0)
    coef = peak_location**(-shape_m1) * np.exp(peak_location / scale)
    return shape_m1 + 1, scale, coef


def gamma_expr(peak_location, peak_fwhm):
    shape, scale, coef = gamma_params(peak_location, peak_fwhm)
    return (
        coef * ((T >= 0) * (T+1.0e-14))**(shape-1)
        * sympy.exp(-(T+1.0e-14)/scale)
        )


def _get_sym_int(f, dt=0.02, t=50):
    # numerical integral of symbolic function
    return _get_num_int(lambdify_t(f), dt, t)


def _get_num_int(lf, dt=0.02, t=50):
    # numerical integral of numerical function
    tt = np.arange(dt,t+dt,dt)
    return lf(tt).sum() * dt


# Glover HRF
_gexpr = gamma_expr(5.4, 5.2) - 0.35 * gamma_expr(10.8, 7.35)
_gexpr = _gexpr / _get_sym_int(_gexpr)
# The numerical function (pass times to get values)
glovert = lambdify_t(_gexpr)
# The symbolic function
glover = implemented_function('glover', glovert)

# Derivative of Glover HRF
_dgexpr = _gexpr.diff(T)
_dpos = sympy.Derivative((T >= 0), T)
_dgexpr = _dgexpr.subs(_dpos, 0)
_dgexpr = _dgexpr / _get_sym_int(sympy.Abs(_dgexpr))
# Numerical function
dglovert = lambdify_t(_dgexpr)
# Symbolic function
dglover = implemented_function('dglover', dglovert)

del(_gexpr); del(_dpos); del(_dgexpr)

# AFNI's HRF
_aexpr = ((T >= 0) * T)**8.6 * sympy.exp(-T/0.547)
_aexpr = _aexpr / _get_sym_int(_aexpr)
# Numerical function
afnit = lambdify_t(_aexpr)
# Symbolic function
afni = implemented_function('afni', afnit)

del(_aexpr)

# SPMs HRF
def spm_hrf_compat(t,
                   peak_delay=6,
                   under_delay=16,
                   peak_disp=1,
                   under_disp=1,
                   p_u_ratio = 6,
                   normalize=True,
                  ):
    """ SPM HRF function from sum of two gamma PDFs

    This function is designed to be partially compatible with SPMs `spm_hrf.m`
    function.

    The SPN HRF is a *peak* gamma PDF (with location `peak_delay` and dispersion
    `peak_disp`), minus an *undershoot* gamma PDF (with location `under_delay`
    and dispersion `under_disp`, and divided by the `p_u_ratio`).

    Parameters
    ----------
    t : array-like
        vector of times at which to sample HRF.
    peak_delay : float, optional
        delay of peak.
    under_delay : float, optional
        delay of undershoot.
    peak_disp : float, optional
        width (dispersion) of peak.
    under_disp : float, optional
        width (dispersion) of undershoot.
    p_u_ratio : float, optional
        peak to undershoot ratio.  Undershoot divided by this value before
        subtracting from peak.
    normalize : {True, False}, optional
        If True, divide HRF values by their sum before returning.  SPM does this
        by default.

    Returns
    -------
    hrf : array
        vector length ``len(t)`` of samples from HRF at times `t`.

    Notes
    -----
    See ``spm_hrf.m`` in the SPM distribution.
    """
    if len([v for v in [peak_delay, peak_disp, under_delay, under_disp]
            if v <= 0]):
        raise ValueError("delays and dispersions must be > 0")
    # gamma.pdf only defined for t > 0
    hrf = np.zeros(t.shape, dtype=np.float)
    pos_t = t[t > 0]
    peak = sps.gamma.pdf(pos_t,
                         peak_delay / peak_disp,
                         loc=0,
                         scale = peak_disp)
    undershoot = sps.gamma.pdf(pos_t,
                               under_delay / under_disp,
                               loc=0,
                               scale = under_disp)
    hrf[t > 0] = peak - undershoot / p_u_ratio
    if not normalize:
        return hrf
    return hrf / np.sum(hrf)


_spm_can_int = _get_num_int(partial(spm_hrf_compat, normalize=False))


def spmt(t):
    """ SPM canonical HRF, HRF values for time values `t`

    This is the canonical HRF function as used in SPM
    """
    return spm_hrf_compat(t, normalize=False) / _spm_can_int


def dspmt(t):
    """ SPM canonical HRF derivative, HRF derivative values for time values `t`

    This is the canonical HRF derivative function as used in SPM.

    It is the numerical difference of the HRF sampled at time `t` minus the
    values sampled at time `t` -1
    """
    t = np.asarray(t)
    return spmt(t) - spmt(t - 1)


_spm_dd_func = partial(spm_hrf_compat, normalize=False, peak_disp=1.01)
_spm_dd_func_int = _get_num_int(_spm_dd_func)

def ddspmt(t):
    """ SPM canonical HRF dispersion derivative, values for time values `t`

    This is the canonical HRF dispersion derivative function as used in SPM.

    It is the numerical difference between the HRF sampled at time `t`, and
    values at `t` for another HRF shape with a small change in the peak
    dispersion parameter (``peak_disp`` in func:`spm_hrf_compat`).
    """
    return (spmt(t) - _spm_dd_func(t) / _spm_dd_func_int) / 0.01


spm = implemented_function('spm', spmt)
dspm = implemented_function('dspm', dspmt)
ddspm = implemented_function('ddspm', ddspmt)



""" Testing nibcompat module
"""
from __future__ import absolute_import
import numpy as np

import nibabel as nib

from nibabel.tmpdirs import InTemporaryDirectory

from ..nibcompat import (get_dataobj, get_affine, get_header,
                         get_unscaled_data)

from numpy.testing import assert_array_equal

from nose.tools import (assert_true, assert_false, assert_raises,
                        assert_equal, assert_not_equal)


def test_funcs():
    class OldNib:
        def get_header(self):
            return 1
        def get_affine(self):
            return np.eye(4)
        _data = 3
    class NewNib:
        header = 1
        affine = np.eye(4)
        dataobj = 3
    for img in OldNib(), NewNib():
        assert_equal(get_header(img), 1)
        assert_array_equal(get_affine(img), np.eye(4))
        assert_equal(get_dataobj(img), 3)


def test_unscaled_data():
    shape = (2, 3, 4)
    data = np.random.normal(size=shape)
    with InTemporaryDirectory():
        for img_class, ext, default_offset in (
            # (nib.Spm99AnalyzeImage, '.img', 0),
            # (nib.Spm2AnalyzeImage, '.img', 0),
            (nib.Nifti1Pair, '.img', 0),
            (nib.Nifti1Image, '.nii', 352),
        ):
            img = img_class(data, np.eye(4))
            img.set_data_dtype(np.dtype(np.int16))
            fname = 'test' + ext
            nib.save(img, fname)
            img_back = nib.load(fname)
            header = get_header(img_back)
            dao = get_dataobj(img_back)
            slope = header['scl_slope']
            inter = (0. if not 'scl_inter' in header else header['scl_inter'])
            if np.isnan(slope):
                slope, inter = dao.slope, dao.inter
            data_back = np.array(dao)
            assert_true(np.allclose(data, data_back, atol=slope / 2.))
            header_copy = header.copy()
            header_copy['vox_offset'] = default_offset
            with open(fname, 'rb') as fobj:
                raw_back = header_copy.raw_data_from_fileobj(fobj)
            unscaled = get_unscaled_data(img_back)
            assert_array_equal(unscaled, raw_back)
            assert_true(np.allclose(unscaled * slope + inter, data_back))
            # delete objects to allow file deletion on Windows
            del raw_back, unscaled



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
DESCRIP = 'Calculate and write results for diagnostic screen'
EPILOG = \
'''nipy_diagnose will generate a series of diagnostic images for a 4D
fMRI image volume.  The following images will be generated.  <ext> is
the input filename extension (e.g. '.nii'):

    * components_<label>.png : plots of PCA basis vectors
    * max_<label><ext> : max image
    * mean_<label><ext> : mean image
    * min_<label><ext> : min image
    * pca_<label><ext> : 4D image of PCA component images
    * pcnt_var_<label>.png : percent variance scree plot for PCA
      components
    * std_<label><ext> : standard deviation image
    * tsdiff_<label>.png : time series diagnostic plot

The filenames for the outputs are of the form
<out-path>/<some_prefix><label><file-ext> where <out-path> is the path
specified by the --out-path option, or the path of the input filename;
<some_prefix> are the standard prefixes above, <label> is given by
--out-label, or by the filename of the input image (with path and
extension removed), and <file-ext> is '.png' for graphics, or the
extension of the input filename for volume images.  For example,
specifying only the input filename ``/some/path/fname.img`` will
generate filenames of the form ``/some/path/components_fname.png,
/some/path/max_fname.img`` etc.
'''

import matplotlib
matplotlib.use('Agg')

from nipy.externals.argparse import (ArgumentParser,
                                     RawDescriptionHelpFormatter)
import nipy.algorithms.diagnostics.commands as nadc


def main():
    parser = ArgumentParser(description=DESCRIP,
                            epilog=EPILOG,
                            formatter_class=RawDescriptionHelpFormatter)
    parser.add_argument('filename', type=str,
                        help='4D image filename')
    parser.add_argument('--out-path', type=str,
                        help='path for output image files')
    parser.add_argument('--out-fname-label', type=str,
                        help='mid part of output image filenames')
    parser.add_argument('--ncomponents', type=int, default=10,
                        help='number of PCA components to write')
    parser.add_argument('--time-axis', type=str, default=-1,
                        help='Image axis for time')
    parser.add_argument('--slice-axis', type=str, default=None,
                        help='Image axis for slice')
    # parse the command line
    args = parser.parse_args()
    nadc.diagnose(args)


if __name__ == '__main__':
    main()



from __future__ import absolute_import
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
import numpy as np

from ..coord_tools import (coord_transform, find_cut_coords,
                           find_maxsep_cut_coords)

from numpy.testing import assert_array_equal

def test_coord_transform_trivial():
    sform = np.eye(4)
    x = np.random.random((10,))
    y = np.random.random((10,))
    z = np.random.random((10,))

    x_, y_, z_ = coord_transform(x, y, z, sform)
    np.testing.assert_array_equal(x, x_)
    np.testing.assert_array_equal(y, y_)
    np.testing.assert_array_equal(z, z_)

    sform[:, -1] = 1
    x_, y_, z_ = coord_transform(x, y, z, sform)
    np.testing.assert_array_equal(x+1, x_)
    np.testing.assert_array_equal(y+1, y_)
    np.testing.assert_array_equal(z+1, z_)


def test_find_cut_coords():
    map = np.zeros((100, 100, 100))
    x_map, y_map, z_map = 50, 10, 40
    map[x_map-30:x_map+30, y_map-3:y_map+3, z_map-10:z_map+10] = 1
    x, y, z = find_cut_coords(map, mask=np.ones(map.shape, np.bool))
    np.testing.assert_array_equal(
                        (int(round(x)), int(round(y)), int(round(z))),
                                (x_map, y_map, z_map))


def test_find_maxsep_cut_coords():
    # Test find_maxsep_cut_coords function
    assert_array_equal(
        find_maxsep_cut_coords(np.ones((2, 3, 5)), np.eye(4)), list(range(5)))
    assert_array_equal(
        find_maxsep_cut_coords(np.ones((2, 3, 5)), np.eye(4), threshold=1),
        list(range(5)))
    assert_array_equal(
        find_maxsep_cut_coords(np.ones((2, 3, 4)), np.eye(4), n_cuts=4),
        list(range(4)))
    map_3d = np.ones((2, 3, 5))
    map_3d[:, :, 1] = 0
    assert_array_equal(
        find_maxsep_cut_coords(map_3d, np.eye(4), n_cuts=4), [0, 2, 3, 4])
    map_3d[:, :, 1] = 0.5
    assert_array_equal(
        find_maxsep_cut_coords(map_3d, np.eye(4), n_cuts=4, threshold=0.6),
        [0, 2, 3, 4])



""" Utility functions for returning slice times from number of slices and TR

Slice timing routines in nipy need a vector of slice times.

Slice times are vectors $t_i i = 0 ... N$ of times, one for each slice, where
$t_i% gives the time at which slice number $i$ was acquired, relative to the
beginning of the volume acquisition.

We like these vectors because they are unambiguous; the indices $i$ refer to
positions in space, and the values $t_i$ refer to times.

But, there are many common slice timing regimes for which it's easy to get the
slice times once you know the volume acquisition time (the TR) and the number of
slices.

For example, if you acquired the slices in a simple ascending order, and you
have 10 slices and the TR was 2.0, then the slice times are:

>>> import numpy as np
>>> np.arange(10) / 10.  * 2.0
array([ 0. ,  0.2,  0.4,  0.6,  0.8,  1. ,  1.2,  1.4,  1.6,  1.8])

These are small convenience functions that accept the number of slices and the
TR as input, and return a vector of slice times:

>>> ascending(10, 2.)
array([ 0. ,  0.2,  0.4,  0.6,  0.8,  1. ,  1.2,  1.4,  1.6,  1.8])
"""
from __future__ import division, print_function, absolute_import

import numpy as np

# Dictionary (key, value) == (name, function) for slice timing functions
SLICETIME_FUNCTIONS = {}

def _dec_filldoc(func):
    """ Fill docstring of slice time function
    """
    func._doc_template = func.__doc__
    func.__doc__ = func.__doc__.format(
        **dict(
            name = func.__name__,
            pstr=
"""Note: slice 0 is the first slice in the voxel data block

    Parameters
    ----------
    n_slices : int
        Number of slices in volume
    TR : float
        Time to acquire one full volume

    Returns
    -------
    slice_times : (n_slices,) ndarray
        Vectors $t_i i = 0 ... N$ of times, one for each slice, where $t_i$
        gives the time at which slice number $i$ was acquired, relative to the
        beginning of the volume acquisition.
    """))
    return func


def _dec_register_stf(func):
    """ Register slice time function in module dictionary """
    name = func.__name__
    SLICETIME_FUNCTIONS[name] = func
    if name.startswith('st_'):
        short_name = name[3:]
        if short_name in SLICETIME_FUNCTIONS:
            raise ValueError(
                "Duplicate short / long function name {0}".format(short_name))
        SLICETIME_FUNCTIONS[short_name] = func
    return func


def _dec_stfunc(func):
    return _dec_register_stf(_dec_filldoc(func))


def _derived_func(name, func):
    def derived(n_slices, TR):
        return func(n_slices, TR)
    derived.__name__ = name
    derived.__doc__ = func._doc_template
    return _dec_stfunc(derived)


@_dec_stfunc
def st_01234(n_slices, TR):
    """ Simple ascending slice sequence

    slice 0 first, slice 1 second etc.

    For example, for 5 slices and a TR of 1:

    >>> {name}(5, 1.)
    array([ 0. ,  0.2,  0.4,  0.6,  0.8])

    {pstr}
    """
    return np.arange(n_slices) / n_slices * TR

ascending = _derived_func('ascending', st_01234)


@_dec_stfunc
def st_43210(n_slices, TR):
    """ Simple descending slice sequence

    slice ``n_slices-1`` first, slice ``n_slices - 2`` second etc.

    For example, for 5 slices and a TR of 1:

    >>> {name}(5, 1.)
    array([ 0.8,  0.6,  0.4,  0.2,  0. ])

    {pstr}
    """
    return np.arange(n_slices)[::-1] / n_slices * TR

descending = _derived_func('descending', st_43210)


@_dec_stfunc
def st_02413(n_slices, TR):
    """Ascend alternate every second slice, starting at first slice

    Collect slice 0 first, slice 2 second up to top.  Then return to collect
    slice 1, slice 3 etc.

    For example, for 5 slices and a TR of 1:

    >>> {name}(5, 1.)
    array([ 0. ,  0.6,  0.2,  0.8,  0.4])

    {pstr}
    """
    one_slice = TR / n_slices
    time_to_space = list(range(0, n_slices, 2)) + list(range(1, n_slices, 2))
    space_to_time = np.argsort(time_to_space)
    return space_to_time * one_slice

asc_alt_2 = _derived_func('asc_alt_2', st_02413)


@_dec_stfunc
def st_13024(n_slices, TR):
    """Ascend alternate every second slice, starting at second slice

    Collect slice 1 first, slice 3 second up to top (highest numbered slice).
    Then return to collect slice 0, slice 2 etc.  This order is rare except on
    Siemens acquisitions with an even number of slices.  See
    :func:`st_odd0_even1` for this logic.

    For example, for 5 slices and a TR of 1:

    >>> {name}(5, 1.)
    array([ 0.4,  0. ,  0.6,  0.2,  0.8])

    {pstr}
    """
    one_slice = TR / n_slices
    time_to_space = list(range(1, n_slices, 2)) + list(range(0, n_slices, 2))
    space_to_time = np.argsort(time_to_space)
    return space_to_time * one_slice

asc_alt_2_1 = _derived_func('asc_alt_2_1', st_13024)


@_dec_stfunc
def st_42031(n_slices, TR):
    """Descend alternate every second slice, starting at last slice

    Collect slice (`n_slices` - 1) first, slice (`nslices` - 3) second down to
    bottom (lowest numbered slice).  Then return to collect slice (`n_slices`
    -2), slice (`n_slices` - 4) etc.

    For example, for 5 slices and a TR of 1:

    >>> {name}(5, 1.)
    array([ 0.4,  0.8,  0.2,  0.6,  0. ])

    {pstr}
    """
    return st_02413(n_slices, TR)[::-1]

desc_alt_2 = _derived_func('desc_alt_2', st_42031)


@_dec_stfunc
def st_odd0_even1(n_slices, TR):
    """Ascend alternate starting at slice 0 for odd, slice 1 for even `n_slices`

    Acquisitions with alternating ascending slices from Siemens scanners often
    seem to have this behavior as default - see:
        https://mri.radiology.uiowa.edu/fmri_images.html

    This means we use the :func:`st_02413` algorithm if `n_slices` is odd,
    and the :func:`st_13024` algorithm if `n_slices is even.

    For example, for 4 slices and a TR of 1:

    >>> {name}(4, 1.)
    array([ 0.5 ,  0.  ,  0.75,  0.25])

    5 slices and a TR of 1:

    >>> {name}(5, 1.)
    array([ 0. ,  0.6,  0.2,  0.8,  0.4])

    {pstr}
    """
    if n_slices % 2 == 0:
        return st_13024(n_slices, TR)
    return st_02413(n_slices, TR)

asc_alt_siemens = _derived_func('asc_alt_siemens', st_odd0_even1)


@_dec_stfunc
def st_03142(n_slices, TR):
    """Ascend alternate, where alternation is by half the volume

    Collect slice 0 then slice ``ceil(n_slices / 2.)`` then slice 1 then slice
    ``ceil(nslices / 2.) + 1`` etc.

    For example, for 5 slices and a TR of 1:

    >>> {name}(5, 1.)
    array([ 0. ,  0.4,  0.8,  0.2,  0.6])

    {pstr}
    """
    one_slice = TR / n_slices
    space_to_time = (list(range(0, n_slices, 2)) +
                     list(range(1, n_slices, 2)))
    return np.array(space_to_time) * one_slice

asc_alt_half = _derived_func('asc_alt_half', st_03142)


@_dec_stfunc
def st_41302(n_slices, TR):
    """Descend alternate, where alternation is by half the volume

    Collect slice (n_slices - 1) then slice ``floor(nslices / 2.) - 1`` then slice
    (n_slices - 2) then slice ``floor(nslices / 2.) - 2`` etc.

    For example, for 5 slices and a TR of 1:

    >>> {name}(5, 1.)
    array([ 0.6,  0.2,  0.8,  0.4,  0. ])

    {pstr}
    """
    return st_03142(n_slices, TR)[::-1]

desc_alt_half = _derived_func('desc_alt_half', st_41302)



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
from __future__ import absolute_import, print_function

import os
from distutils import log

try:  # Python 3
    from configparser import ConfigParser, NoSectionError, NoOptionError
except ImportError:  # Python 2
    from ConfigParser import ConfigParser, NoSectionError, NoOptionError

# Global variables
LIBS = os.path.realpath('lib')

# Stuff for reading setup file
SETUP_FILE = 'setup.cfg'
SECTION = 'lapack'
KEY = 'external'
EXTERNAL_LAPACK_VAR = 'NIPY_EXTERNAL_LAPACK'

def get_link_external():
    """ Return True if we should link to system BLAS / LAPACK

    If True, attempt to link to system BLAS / LAPACK.  Otherwise, compile
    lapack_lite, and link to that.

    First check ``setup.cfg`` file for section ``[lapack]`` key ``external``.

    If this value undefined, then get string from environment variable
    NIPY_EXTERNAL_LAPACK.

    If value from ``setup.cfg`` or environment variable is not 'False' or '0',
    then return True.
    """
    config = ConfigParser()
    try:
        config.read(SETUP_FILE)
        external_link = config.get(SECTION, KEY)
    except (IOError, KeyError, NoOptionError, NoSectionError):
        external_link = os.environ.get(EXTERNAL_LAPACK_VAR)
    if external_link is None:
        return False
    return external_link.lower() not in ('0', 'false')


def configuration(parent_package='',top_path=None):
    from numpy.distutils.misc_util import Configuration, get_numpy_include_dirs
    from numpy.distutils.system_info import get_info

    config = Configuration('labs', parent_package, top_path)

    # fff library
    config.add_include_dirs(os.path.join(LIBS,'fff'))
    config.add_include_dirs(os.path.join(LIBS,'fff_python_wrapper'))
    config.add_include_dirs(get_numpy_include_dirs())

    sources = [os.path.join(LIBS,'fff','*.c')]
    sources.append(os.path.join(LIBS,'fff_python_wrapper','*.c'))

    # Link with lapack if found on the system

    # XXX: We need to better sort out the use of get_info() for Lapack, because
    # using 'lapack' and 'lapack_opt' returns different things even comparing
    # Ubuntu 8.10 machines on 32 vs 64 bit setups.  On OSX
    # get_info('lapack_opt') does not return the keys: 'libraries' and
    # 'library_dirs', but get_info('lapack') does.
    #
    # For now this code should do the right thing on OSX and linux, but we
    # should ask on the numpy list for clarification on the proper approach.

    # XXX: If you modify these lines, remember to pass the information
    # along to the different .so in the neurospin build system.
    # First, try 'lapack_info', as that seems to provide more details on Linux
    # (both 32 and 64 bits):
    lapack_info = get_info('lapack_opt', 0)
    if 'libraries' not in lapack_info:
        # But on OSX that may not give us what we need, so try with 'lapack'
        # instead.  NOTE: scipy.linalg uses lapack_opt, not 'lapack'...
        lapack_info = get_info('lapack', 0)

    # If lapack linking not required or no lapack install is found, we
    # use the rescue lapack lite distribution included in the package
    # (sources have been translated to C using f2c)
    want_lapack_link = get_link_external()
    if not want_lapack_link:
        log.warn('Building with (slow) Lapack lite distribution: '
                 'set {0} environment variable or use setup.cfg '
                 'to enable link to optimized BLAS / LAPACK'.format(
                        EXTERNAL_LAPACK_VAR)
                 )
        sources.append(os.path.join(LIBS,'lapack_lite','*.c'))
        library_dirs = []
        libraries = []
    else: # Best-case scenario: external lapack found
        if not lapack_info:
            raise RuntimeError('Specified external lapack linking but '
                               'numpy does not report external lapack')
        log.warn('Linking with system Lapack')
        library_dirs = lapack_info['library_dirs']
        libraries = lapack_info['libraries']
        if 'include_dirs' in lapack_info:
            config.add_include_dirs(lapack_info['include_dirs'])

    # Information message
    print('LAPACK build options:')
    print('library_dirs: %s ' % library_dirs)
    print('libraries: %s ' % libraries)
    print('lapack_info: %s ' % lapack_info)

    config.add_library('cstat',
                       sources=sources,
                       library_dirs=library_dirs,
                       libraries=libraries,
                       extra_info=lapack_info)

    # Subpackages
    config.add_subpackage('bindings')
    config.add_subpackage('glm')
    config.add_subpackage('group')
    config.add_subpackage('spatial_models')
    config.add_subpackage('utils')
    config.add_subpackage('viz_tools')
    config.add_subpackage('datasets')
    config.add_subpackage('tests')

    config.make_config_py() # installs __config__.py

    return config

if __name__ == '__main__':
    print('This is the wrong setup.py file to run')



#!/usr/bin/env python
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
from __future__ import absolute_import

from os.path import split as psplit

import numpy as np

import nibabel.eulerangles as euler

from nipy.io.api import load_image, save_image
import nipy.algorithms.registration as reg

from nose.tools import assert_true, assert_false

import numpy.testing as npt

from nipy.testing import funcfile

from nibabel.tmpdirs import InTemporaryDirectory


def test_space_time_realign():
    path, fname = psplit(funcfile)
    original_affine = load_image(funcfile).affine
    path, fname = psplit(funcfile)
    froot, _ = fname.split('.', 1)
    with InTemporaryDirectory():
        # Make another image with .nii extension and extra dot in filename
        save_image(load_image(funcfile), 'my.test.nii')
        for in_fname, out_fname in ((funcfile, froot + '_mc.nii.gz'),
                                    ('my.test.nii', 'my.test_mc.nii.gz')):
            xforms = reg.space_time_realign(in_fname, 2.0, out_name='.')
            assert_true(np.allclose(xforms[0].as_affine(), np.eye(4), atol=1e-7))
            assert_false(np.allclose(xforms[-1].as_affine(), np.eye(4), atol=1e-3))
            img = load_image(out_fname)
            npt.assert_almost_equal(original_affine, img.affine)


def test_aff2euler():
    xr = 0.1
    yr = -1.3
    zr = 3.1
    scales = (2.1, 3.2, 4.4)
    R = np.dot(euler.euler2mat(xr, yr, zr), np.diag(scales))
    aff = np.eye(4)
    aff[:3, :3] = R
    aff[:3, 3] = [11, 12, 13]
    npt.assert_almost_equal(reg.aff2euler(aff), (xr, yr, zr))



from __future__ import division
from __future__ import absolute_import

import numpy as np

from scipy.ndimage import gaussian_filter, gaussian_filter1d

from nose.tools import assert_true, assert_false
from numpy.testing import assert_almost_equal, assert_array_equal

from nipy.core.api import Image, vox2scanner
from ..groupwise_registration import SpaceTimeRealign


def check_stc(true_signal, corrected_signal, ref_slice=0,
              rtol=1e-5, atol=1e-5):
    n_slices = true_signal.shape[2]
    # The reference slice should be more or less perfect
    assert_almost_equal(
        corrected_signal[..., ref_slice, :],
        true_signal[..., ref_slice, :])
    # The other slices should be more or less right
    for sno in range(n_slices):
        if sno == ref_slice:
            continue # We checked this one
        arr0 = true_signal[..., sno, 1:-1]
        arr1 = corrected_signal[..., sno, 1:-1]
        # Intermediate test matrices for debugging
        abs_diff = np.abs(arr0 - arr1)
        rel_diff = np.abs((arr0 / arr1) - 1)
        abs_fails = abs_diff > atol
        rel_fails = rel_diff > rtol
        fails = abs_fails & rel_fails
        abs_only = abs_diff[fails]
        rel_only = rel_diff[fails]
        assert_true(np.allclose(arr0, arr1, rtol=rtol, atol=atol))


def test_slice_time_correction():
    # Make smooth time course at slice resolution
    TR = 2.
    n_vols = 25
    n_slices = 10
    # Create single volume
    shape_3d = (20, 30, n_slices)
    spatial_sigma = 4
    time_sigma = n_slices * 5 # time sigma in TRs
    one_vol = np.random.normal(100, 25, size=shape_3d)
    gaussian_filter(one_vol, spatial_sigma, output=one_vol)
    # Add smoothed time courses.  Time courses are at time resolution of one
    # slice time.  So, there are n_slices time points per TR.
    n_vol_slices = n_slices * n_vols
    time_courses = np.random.normal(0, 15, size=shape_3d + (n_vol_slices,))
    gaussian_filter1d(time_courses, time_sigma, output=time_courses)
    big_data = one_vol[..., None] + time_courses
    # Can the first time point be approximated from the later ones?
    first_signal = big_data[..., 0:n_vol_slices:n_slices]
    for name, time_to_slice in (
        ('ascending', list(range(n_slices))),
        ('descending', list(range(n_slices)[::-1])),
        ('asc_alt_2', (list(range(0, n_slices, 2)) +
                       list(range(1, n_slices, 2)))),
        ('desc_alt_2', (list(range(0, n_slices, 2)) +
                        list(range(1, n_slices, 2)))[::-1])
    ):
        slice_to_time = np.argsort(time_to_slice)
        acquired_signal = np.zeros_like(first_signal)
        for space_sno, time_sno in enumerate(slice_to_time):
            acquired_signal[..., space_sno, :] = \
                big_data[..., space_sno, time_sno:n_vol_slices:n_slices]
        # do STC - minimizer will fail
        acquired_image = Image(acquired_signal, vox2scanner(np.eye(5)))
        stc = SpaceTimeRealign(acquired_image, TR, name, 2)
        stc.estimate(refscan=None, loops=1, between_loops=1, optimizer='steepest')
        # Check no motion estimated
        assert_array_equal([t.param for t in stc._transforms[0]], 0)
        corrected = stc.resample()[0].get_data()
        # check we approximate first time slice with correction
        assert_false(np.allclose(acquired_signal, corrected, rtol=1e-3,
                                 atol=0.1))
        check_stc(first_signal, corrected, ref_slice=slice_to_time[0],
                  rtol=5e-4, atol=1e-6)



print("import: 'nipy'")
import nipy

print("import: 'nipy.algorithms'")
import nipy.algorithms

print("import: 'nipy.core'")
import nipy.core

print("import: 'nipy.interfaces'")
import nipy.interfaces

print("import: 'nipy.io'")
import nipy.io

print("import: 'nipy.labs'")
import nipy.labs

print("import: 'nipy.modalities'")
import nipy.modalities




# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:

"""This is a wrapper of SpaceTimeRealign

Based on:

Alexis Roche (2011) A Four-Dimensional Registration Algorithm With Application to Joint Correction of Motion and Slice Timing in fMRI. IEEE Trans. Med. Imaging 30(8): 1546-1554

"""
import os
import os.path as op
import nipy.algorithms.registration as reg
import nipy.externals.argparse as argparse

parser = argparse.ArgumentParser()

parser.add_argument('TR', type=float, metavar='Float', help="""The TR of the measurement""")

parser.add_argument('input', type=str, metavar='File',
                help="""Path to a nifti file, or to a folder containing nifti files. If a path to a folder is provided, the order of motion correction will be np.sort(list_of_files). The outputs will be '*_mc.par' (containing 3 translation and three rotation parameters) and '*_mc.nii.gz' containing the motion corrected data (unless 'apply' is set to False)""")

parser.add_argument('--slice_order', type=str, metavar='String',
                    help="""The order of slice aquisition {'ascending', 'descending' (default), or the name of a function from `nipy.algorithms.slicetiming.timefuncs`}""", default='descending')

parser.add_argument('--slice_dim', type=int, metavar='Int', help="""Integer
denoting the axis in `images` that is the slice axis.  In a 4D image, this will
often be axis = 2 (default).""", default=2)

parser.add_argument('--slice_dir', type=int, metavar='Int', help=""" 1 if the
slices were acquired slice 0 first (default), slice -1 last, or -1 if acquire slice -1 first, slice 0 last.""", default=1)

parser.add_argument('--make_figure', type=bool, metavar='Bool',
                help="""Whether to generate a '.png' figure with the motion parameters across runs. {True, False}. Default: False """, default=False)

parser.add_argument('--save_path', type=str, metavar='String',
                    help="""Full path to a file-system location for the output files. Defaults to the same location as the input files""",
                    default='none')

parser.add_argument('--save_params', type=bool, metavar='Bool',
                help="""Whether to save the motion corrections parameters (3 rotations, 3 translations). {True, False}. Default: False. NOTE: The rotations are not Euler angles, but a rotation vector. Use `nipy.algorithms.registration.to_matrix44` to convert to a 4-by-4 affine matrix""", default=False)


# parse the command line
args = parser.parse_args()

if __name__ == '__main__':
    if args.save_path == 'none':
        save_path = op.split(args.input)[0]
    else:
        save_path = args.save_path
        
    xform = reg.space_time_realign(args.input, float(args.TR),
                               slice_order=args.slice_order,
                               slice_dim=int(args.slice_dim),
                               slice_dir=int(args.slice_dir),
                               apply=True, # We always apply the xform in the cli
                               make_figure=args.make_figure,
                               out_name=save_path)

    if args.save_params:
        
        f = file(op.join(save_path, 'mc.par'), 'w')
        for x in xform:
            euler_rot = reg.aff2euler(x.as_affine())
            for r in euler_rot:
                f.write('%s\t'%r)
            for t in x.translation[:-1]:
                f.write('%s\t'%t)
            f.write('%s\n'%x.translation[-1])
        f.close()



""" Testing utilities module
"""
from __future__ import absolute_import

import numpy as np

from nose.tools import (assert_true, assert_false, assert_raises,
                        assert_equal, assert_not_equal)


from ..utilities import is_iterable, is_numlike, seq_prod

def test_is_iterable():
    assert_true(is_iterable(()))
    assert_true(is_iterable([]))
    assert_true(is_iterable(np.zeros(1)))
    assert_true(is_iterable(np.zeros((1, 1))))
    assert_true(is_iterable(''))
    assert_false(is_iterable(0))
    assert_false(is_iterable(object()))

    def gen():
        yield 1

    assert_true(is_iterable(gen()))

    def func():
        return 1

    assert_false(is_iterable(func))

    class C:
        def __iter__(self):
            return self

        def __next__(self):
            return self

    assert_true(is_iterable(C()))


def test_is_numlike():
    for good in (1, 0, 1.1, False, True, np.zeros(1), np.zeros((3,)),
                 1j, np.complex128(1)):
        assert_true(is_numlike(good))
    for bad in ('', object(), np.array(''), [], [1], (), (1,)):
        assert_false(is_numlike(bad))


def test_seq_prod():
    assert_equal(seq_prod(()), 1)
    assert_equal(seq_prod((1,)), 1)
    assert_equal(seq_prod((1, 2)), 2)
    assert_equal(seq_prod((1, 2), 2), 4)
    assert_equal(seq_prod((1, 2), 2.), 4.)



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
from __future__ import absolute_import

from warnings import warn

from . import onesample
from . import twosample
from . import glm_twolevel
from . import permutation_test

warn('Module nipy.labs.group deprecated, will be removed',
     FutureWarning,
     stacklevel=2)

from nipy.testing import Tester
test = Tester().test
bench = Tester().bench



""" This file contains defines parameters for nipy that we use to fill
settings in setup.py, the nipy top-level docstring, and for building the
docs.  In setup.py in particular, we exec this file, so it cannot import nipy
"""

# nipy version information.  An empty _version_extra corresponds to a
# full release.  '.dev' as a _version_extra string means this is a development
# version
_version_major = 0
_version_minor = 4
_version_micro = 1
#_version_extra = '.dev' # For development
_version_extra = '' # For release

# Format expected by setup.py and doc/source/conf.py: string of form "X.Y.Z"
__version__ = "%s.%s.%s%s" % (_version_major,
                              _version_minor,
                              _version_micro,
                              _version_extra)

CLASSIFIERS = ["Development Status :: 3 - Alpha",
               "Environment :: Console",
               "Intended Audience :: Science/Research",
               "License :: OSI Approved :: BSD License",
               "Operating System :: OS Independent",
               "Programming Language :: Python",
               "Topic :: Scientific/Engineering"]

description  = 'A python package for analysis of neuroimaging data'

# Note: this long_description is the canonical place to edit this text.
# It also appears in README.rst, but it should get there by running
# ``tools/refresh_readme.py`` which pulls in this version.
long_description = \
"""
====
NIPY
====

Neuroimaging tools for Python.

The aim of NIPY is to produce a platform-independent Python environment for
the analysis of functional brain imaging data using an open development model.

In NIPY we aim to:

1. Provide an open source, mixed language scientific programming environment
   suitable for rapid development.

2. Create software components in this environment to make it easy to develop
   tools for MRI, EEG, PET and other modalities.

3. Create and maintain a wide base of developers to contribute to this
   platform.

4. To maintain and develop this framework as a single, easily installable
   bundle.

NIPY is the work of many people. We list the main authors in the file
``AUTHOR`` in the NIPY distribution, and other contributions in ``THANKS``.

Website
=======

Current information can always be found at the `NIPY project website
<http://nipy.org/nipy>`_.

Mailing Lists
=============

For questions on how to use nipy or on making code contributions, please see
the ``neuroimaging`` mailing list:

    https://mail.python.org/mailman/listinfo/neuroimaging

Please report bugs at github issues:

    https://github.com/nipy/nipy/issues

You can see the list of current proposed changes at:

    https://github.com/nipy/nipy/pulls

Code
====

You can find our sources and single-click downloads:

* `Main repository`_ on Github;
* Documentation_ for all releases and current development tree;
* Download the `current development version`_ as a tar/zip file;
* Downloads of all `available releases`_.

.. _main repository: https://github.com/nipy/nipy
.. _Documentation: http://nipy.org/nipy
.. _current development version: https://github.com/nipy/nipy/archive/master.zip
.. _available releases: http://pypi.python.org/pypi/nipy

Tests
=====

To run nipy's tests, you will need to install the nose_ Python testing
package.  If you are using Python 2.7, you will also need to install the mock_
testing package - e.g.::

    pip install nose mock

Then::

    python -c "import nipy; nipy.test()"

You can also run nipy's tests with the ``nipnost`` script in the ``tools``
directory of the nipy distribution::

    ./tools/nipnost nipy

``nipnost`` is a thin wrapper around the standard ``nosetests`` program that
is part of the nose package.  Try ``nipnost --help`` to see a large number of
command-line options.

Dependencies
============

To run NIPY, you will need:

* python_ >= 2.6 (tested with 2.6, 2.7, 3.2 through 3.5)
* numpy_ >= 1.6.0
* scipy_ >= 0.9.0
* sympy_ >= 0.7.0
* nibabel_ >= 1.2

You will probably also like to have:

* ipython_ for interactive work
* matplotlib_ for 2D plotting
* mayavi_ for 3D plotting

.. _python: http://python.org
.. _numpy: http://numpy.scipy.org
.. _scipy: http://www.scipy.org
.. _sympy: http://sympy.org
.. _nibabel: http://nipy.org/nibabel
.. _ipython: http://ipython.org
.. _matplotlib: http://matplotlib.org
.. _mayavi: http://code.enthought.com/projects/mayavi/
.. _nose: http://nose.readthedocs.org/en/latest
.. _mock: https://pypi.python.org/pypi/mock

License
=======

We use the 3-clause BSD license; the full license is in the file ``LICENSE``
in the nipy distribution.
"""

# minimum versions
# Update in readme text above
# Update in .travis.yml
# Update in requirements.txt
NUMPY_MIN_VERSION='1.6.0'
SCIPY_MIN_VERSION = '0.9.0'
NIBABEL_MIN_VERSION = '1.2'
SYMPY_MIN_VERSION = '0.7.0'
MAYAVI_MIN_VERSION = '3.0'
CYTHON_MIN_VERSION = '0.12.1'

NAME                = 'nipy'
MAINTAINER          = "nipy developers"
MAINTAINER_EMAIL    = "neuroimaging@python.org"
DESCRIPTION         = description
LONG_DESCRIPTION    = long_description
URL                 = "http://nipy.org/nipy"
DOWNLOAD_URL        = "http://github.com/nipy/nipy/archives/master"
LICENSE             = "BSD license"
CLASSIFIERS         = CLASSIFIERS
AUTHOR              = "nipy developmers"
AUTHOR_EMAIL        = "neuroimaging@python.org"
PLATFORMS           = "OS Independent"
MAJOR               = _version_major
MINOR               = _version_minor
MICRO               = _version_micro
ISRELEASE           = _version_extra == ''
VERSION             = __version__
REQUIRES            = ["numpy", "scipy", "sympy", "nibabel"]
STATUS              = 'beta'

# Versions and locations of optional data packages
NIPY_DATA_URL= 'http://nipy.org/data-packages/'
DATA_PKGS = {'nipy-data': {'min version':'0.2',
                           'relpath':'nipy/data'},
             'nipy-templates': {'min version':'0.2',
                                'relpath':'nipy/templates'}
            }
NIPY_INSTALL_HINT = \
"""You can download and install the package from:

%s

Check the instructions in the ``doc/users/install_data.rst`` file in the nipy
source tree, or online at http://nipy.org/nipy/users/install_data.html

If you have the package, have you set the path to the package correctly?"""

for key, value in DATA_PKGS.items():
    url = '%s%s-%s.tar.gz' % (NIPY_DATA_URL,
                              key,
                              value['min version'])
    value['name'] = key
    value['install hint'] = NIPY_INSTALL_HINT % url

del key, value, url




# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Conversion mechansims for IO and interaction between volumetric datasets 
and other type of neuroimaging data.
"""
from __future__ import absolute_import
import os

import numpy as np

import nibabel as nib
from nibabel.spatialimages import SpatialImage

from nipy.io.nibcompat import get_header, get_affine
from .volumes.volume_img import VolumeImg

from nipy.externals.six import string_types

def as_volume_img(obj, copy=True, squeeze=True, world_space=None):
    """ Convert the input to a VolumeImg.

        Parameters
        ----------
        obj : filename, pynifti or brifti object, or volume dataset.
            Input object, in any form that can be converted to a
            VolumeImg. This includes Nifti filenames, pynifti or brifti
            objects, or other volumetric dataset objects.
        copy: boolean, optional
            If copy is True, the data and affine arrays are copied, 
            elsewhere a view is taken.
        squeeze: boolean, optional
            If squeeze is True, the data array is squeeze on for
            dimensions above 3.
        world_space: string or None, optional
            An optional specification of the world space, to override
            that given by the image.

        Returns
        -------
        volume_img: VolumeImg object
            A VolumeImg object containing the data. The metadata is
            kept as much as possible in the metadata attribute.

        Notes
        ------
        The world space might not be correctly defined by the input
        object (in particular, when loading data from disk). In this
        case, you can correct it manually using the world_space keyword
        argument.

        For pynifti objects, the data is transposed.
    """
    if hasattr(obj, 'as_volume_img'):
        obj = obj.as_volume_img(copy=copy)
        if copy:
            obj = obj.__copy__()
        return obj

    elif isinstance(obj, string_types):
        if not os.path.exists(obj):
            raise ValueError("The file '%s' cannot be found" % obj)
        obj = nib.load(obj)
        copy = False

    if isinstance(obj, SpatialImage):
        data   = obj.get_data()
        affine = get_affine(obj)
        header = dict(get_header(obj))
        fname = obj.file_map['image'].filename
        if fname:
            header['filename'] = fname
    elif hasattr(obj, 'data') and hasattr(obj, 'sform') and \
                                            hasattr(obj, 'getVolumeExtent'):
        # Duck-types to a pynifti object
        data     = obj.data.T
        affine   = obj.sform
        header   = obj.header
        filename = obj.getFilename()
        if filename != '':
            header['filename'] = filename
    else:
        raise ValueError('Invalid type (%s) passed in: cannot convert %s to '
                    'VolumeImg' % (type(obj), obj))

    if world_space is None and header.get('sform_code', 0) == 4:
        world_space = 'mni152'

    data    = np.asanyarray(data)
    affine  = np.asanyarray(affine)
    if copy:
        data    = data.copy()
        affine  = affine.copy()

    if squeeze:
        # Squeeze the dimensions above 3
        shape = [val for index, val in enumerate(data.shape)
                     if val !=1 or index < 3]
        data = np.reshape(data, shape)
    
    return VolumeImg(data, affine, world_space, metadata=header)


def save(filename, obj):
    """ Save an nipy image object to a file.
    """
    obj = as_volume_img(obj, copy=False)
    hdr = nib.Nifti1Header()
    for key, value in obj.metadata.items():
        if key in hdr:
            hdr[key] = value
    img = nib.Nifti1Image(obj.get_data(),
                          obj.affine,
                          header=hdr)
    nib.save(img, filename)




# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
DESCRIP = ' Analyze, plot time series difference metrics'
EPILOG = \
'''nipy_tsdiffana runs the time series diference algorithm over a 4D
image volume, often and FMRI volume.

It works in one of three modes:

* interactive : the time series difference plot appears on screen.  This is the
  default mode
* non-interactive, plot only : write time series difference plot to graphic
  file. Use the "--out-file=<myfilename>" option to activate this mode
* non-interactive, write plot, images and variables : write plot to file, and
  write generated diagnostic images and variables to files as well. Use the
  "--write-results" flag to activate this option. The generated filenames come
  from the results of the "--out-path" and "--out-fname-label" options (see
  help).

Write-results option, generated files
-------------------------------------

When doing the time point analysis, we will make a difference volume between
each time point and the next time point in the series.  If we have T volumes
then there will be (T-1) difference volumes. Call the vector of difference
volumes DV and the first difference volume DV[0].  So DV[0] results from
subtraction of the second volume in the 4D input image from the first volume in
the 4D input image.  The element-wise squared values from DV[0] is *DV2[0]*.

The following images will be generated. <ext> is the input filename extension
(e.g. '.nii'):

* "dv2_max_<label><ext>" : 3D image volume, where each slice S is slice from
  all of DV2[0] (slice S) throudh DV2[T-1] (slice S) that has the maximum
  summed squared values.  This volume gives an idea of the worst (highest
  difference) slices across the whole time series.
* "dv2_mean_<label><ext>" : the mean of all DV2 volumes DV2[0] .. DV[T-1]
  across the volume (time) dimension.  Higher voxel values in this volume mean
  that time-point to time point differences tended to be high in this voxel.

We also write the mean signal at each time point, and the mean squared
difference between each slice in time, as variables to a 'npz' file named
"tsdiff_<label>.npz"

The filenames for the outputs are of the form
<out-path>/<some_prefix><label><file-ext> where <out-path> is the path
specified by the --out-path option, or the path of the input filename;
<some_prefix> is one of the standard prefixes above, <label> is given by
--out-label, or by the filename of the input image (with path and extension
removed), and <file-ext> is '.png' for graphics, or the extension of the input
filename for volume images.  For example, specifying only the input filename
``/some/path/fname.img`` will generate filenames of the form
``/some/path/tsdiff_fname.png, /some/path/dv2_max_fname.img`` etc.
'''

from nipy.externals.argparse import (ArgumentParser,
                                     RawDescriptionHelpFormatter)


def main():
    parser = ArgumentParser(description=DESCRIP,
                            epilog=EPILOG,
                            formatter_class=RawDescriptionHelpFormatter)
    # add the arguments
    parser.add_argument('filename', type=str,
                        help='4D image filename')
    parser.add_argument('--out-file', type=str,
                        help='graphics file to write to instead '
                        'of leaving image on screen')
    parser.add_argument('--write-results', action='store_true',
                        help='if specified, write diagnostic images and '
                        'analysis variables, plot to OUT_PATH. Mutually '
                        'incompatible with OUT_FILE')
    parser.add_argument('--out-path', type=str,
                        help='path for output image files (default from '
                        'FILENAME path')
    parser.add_argument('--out-fname-label', type=str,
                        help='mid part of output image / plot filenames')
    parser.add_argument('--time-axis', type=str, default='t',
                        help='Image axis for time')
    parser.add_argument('--slice-axis', type=str, default=None,
                        help='Image axis for slice')
    # parse the command line
    args = parser.parse_args()
    if args.out_file is not None and args.write_results == True:
        raise RuntimeError("OUT_FILE option not compatible with WRITE_RESULTS"
                           " option")
    show = args.out_file is None and args.write_results == False
    if not show:
        import matplotlib
        matplotlib.use('Agg')
    else:
        import matplotlib.pylab as plt
    # Import late to give set of mpl backend best chance of working
    from nipy.algorithms.diagnostics.commands import tsdiffana
    tsdiffana(args)
    if show:
        plt.show()


if __name__ == '__main__':
    main()



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
DESCRIP = 'Read 3D image files and write a 4D file'
EPILOG = \
'''nipy_3dto4d will take a series of 3D nifti images in any format readable by
nibabel and concatenate them into a 4D image, and write the image with format
guessed from the output image filename. You can set the filename with the
``--out-4d`` parameter, or we make a filename from the input names.
'''

import os
from os.path import join as pjoin

from nipy.externals.argparse import (ArgumentParser,
                                     RawDescriptionHelpFormatter)

import nibabel as nib


def do_3d_to_4d(filenames, check_affines=True):
    imgs = []
    for fname in filenames:
        img = nib.load(fname)
        imgs.append(img)
    return nib.concat_images(imgs, check_affines=check_affines)


def main():
    parser = ArgumentParser(description=DESCRIP,
                            epilog=EPILOG,
                            formatter_class=RawDescriptionHelpFormatter)
    parser.add_argument('in_filenames', type=str,
                        nargs='+',
                        help='3D image filenames')
    parser.add_argument('--out-4d', type=str,
                        help='4D output image name')
    parser.add_argument('--check-affines', type=bool,
                        default=True,
                        help='False if you want to ignore differences '
                        'in affines between the 3D images, True if you '
                        'want to raise an error for significant '
                        'differences (default is True)')
    # parse the command line
    args = parser.parse_args()
    # get input 3ds
    filenames = args.in_filenames
    # affine check
    check_affines = args.check_affines
    # get output name
    out_fname = args.out_4d
    if out_fname is None:
        pth, fname = os.path.split(filenames[0])
        froot, ext = os.path.splitext(fname)
        if ext in ('.gz', '.bz2'):
            gz = ext
            froot, ext = os.path.splitext(froot)
        else:
            gz = ''
        out_fname = pjoin(pth, froot + '_4d' + ext + gz)
    img4d = do_3d_to_4d(filenames, check_affines=check_affines)
    nib.save(img4d, out_fname)


if __name__ == '__main__':
    main()



""" Patches for scipy.ndimage

Patched affine_transform to work round np.intp bug.

Some versions of scipy.ndimage interpolation routines can't handle the np.intp
type. I (MB) have only seen this on a 32-bit machine runing scipy 0.9.0
"""

import numpy as np

import scipy.ndimage as spnd

# int dtype corresponding to intp
_INT_DTYPE = np.dtype(np.int32 if np.dtype(np.intp).itemsize == 4 else
                      np.int64)


# Do not run doctests on this module via nose; scipy doctests unreliable
__test__ = False


def _proc_array(array):
    """ Change array dtype from intp to int32 / int64

    Parameters
    ----------
    array : ndarray

    Returns
    -------
    output_array : ndarray
        `array` unchanged or view of array where array dtype has been changed
        from ``np.intp`` to ``np.int32`` or ``np.int64`` depending on whether
        this is a 32 or 64 bit numpy.  All other dtypes unchanged.
    """
    if array.dtype == np.dtype(np.intp):
        return array.view(_INT_DTYPE)
    return array


def _proc_output(output):
    """ Change dtype from intp to int32 / int64 for ndimage output parameter

    Allowed values to `output` are the same as listed for the ``output``
    parameter to ``scipy.ndimage.affine_transform``.

    Parameters
    ----------
    output : None or ndarray or or dtype or dtype specifier
        Can be ndarray (will have ``.dtype`` attribute), or a numpy dtype, or
        something that can be converted to a numpy dtype such as a string dtype
        code or numpy type.

    Returns
    -------
    output_fixed : None or ndarray or dtype or dtype specifier
        `output` where array dtype or dtype specifier has been changed from
        ``np.intp`` to ``np.int32`` or ``np.int64`` depending on whether this
        is a 32 or 64 bit numpy.  All other dtypes unchanged.  None returned
        unchanged
    """
    if output is None:
        return None
    if hasattr(output, 'dtype'):  # output was ndarray
        return _proc_array(output)
    # output can also be a dtype specifier
    if np.dtype(output) == np.dtype(np.intp):  # dtype specifier for np.intp
        return _INT_DTYPE
    return output


def affine_transform(input, matrix, offset=0.0, output_shape=None, output=None,
                     order=3, mode='constant', cval=0.0, prefilter=True):
    return spnd.affine_transform(_proc_array(input), matrix, offset,
                                 output_shape, _proc_output(output), order,
                                 mode, cval, prefilter)

affine_transform.__doc__ = spnd.affine_transform.__doc__


def map_coordinates(input, coordinates, output=None, order=3,
                    mode='constant', cval=0.0, prefilter=True):
    return spnd.map_coordinates(_proc_array(input), coordinates,
                                _proc_output(output), order, mode, cval,
                                prefilter)

map_coordinates.__doc__ = spnd.map_coordinates.__doc__



from __future__ import absolute_import

from os.path import exists

import numpy as np

from nibabel import Nifti1Image

from ...utils.simul_multisubject_fmri_dataset import surrogate_3d_dataset
from ..bsa_io import make_bsa_image

from nose.tools import assert_true
from numpy.testing import assert_equal

from nibabel.tmpdirs import InTemporaryDirectory


def test_parcel_intra_from_3d_images_list():
    """Test that a parcellation is generated, starting from a list of 3D images
    """
    # Generate an image
    shape = (5, 5, 5)
    contrast_id = 'plop'
    mask_image = Nifti1Image(np.ones(shape), np.eye(4))
    #mask_images = [mask_image for _ in range(5)]

    with InTemporaryDirectory() as dir_context:
        data_image = ['image_%d.nii' % i for i in range(5)]
        for datim in data_image:
            surrogate_3d_dataset(mask=mask_image, out_image_file=datim)

        #run the algo
        landmark, hrois = make_bsa_image(
            mask_image, data_image, threshold=10., smin=0, sigma=1.,
            prevalence_threshold=0, prevalence_pval=0.5, write_dir=dir_context,
            algorithm='density', contrast_id=contrast_id)

        assert_equal(landmark, None)
        assert_equal(len(hrois), 5)
        assert_true(exists('density_%s.nii' % contrast_id))
        assert_true(exists('prevalence_%s.nii' % contrast_id))
        assert_true(exists('AR_%s.nii' % contrast_id))
        assert_true(exists('CR_%s.nii' % contrast_id))


if __name__ == "__main__":
    import nose
    nose.run(argv=['', __file__])



from __future__ import absolute_import
from .brain_segmentation import BrainT1Segmentation
from .segmentation import Segmentation, moment_matching

from nipy.testing import Tester
test = Tester().test
bench = Tester().bench



from __future__ import absolute_import
from ._registration import _L1_moments

import numpy as np
from scipy.ndimage import gaussian_filter

TINY = float(np.finfo(np.double).tiny)
SIGMA_FACTOR = 0.05

# A lambda function to force positive values
nonzero = lambda x: np.maximum(x, TINY)


def correlation2loglikelihood(rho2, npts):
    """
    Re-normalize correlation.

    Convert a squared normalized correlation to a proper
    log-likelihood associated with a registration problem. The result
    is a function of both the input correlation and the number of
    points in the image overlap.

    See: Roche, medical image registration through statistical
    inference, 2001.

    Parameters
    ----------
    rho2: float
      Squared correlation measure

    npts: int
      Number of points involved in computing `rho2`

    Returns
    -------
    ll: float
      Log-likelihood re-normalized `rho2`
    """
    return -.5 * npts * np.log(nonzero(1 - rho2))


def dist2loss(q, qI=None, qJ=None):
    """
    Convert a joint distribution model q(i,j) into a pointwise loss:

    L(i,j) = - log q(i,j)/(q(i)q(j))

    where q(i) = sum_j q(i,j) and q(j) = sum_i q(i,j)

    See: Roche, medical image registration through statistical
    inference, 2001.
    """
    qT = q.T
    if qI is None:
        qI = q.sum(0)
    if qJ is None:
        qJ = q.sum(1)
    q /= nonzero(qI)
    qT /= nonzero(qJ)
    return -np.log(nonzero(q))


class SimilarityMeasure(object):
    """
    Template class
    """
    def __init__(self, shape, renormalize=False, dist=None):
        self.shape = shape
        self.J, self.I = np.indices(shape)
        self.renormalize = renormalize
        if dist is None:
            self.dist = None
        else:
            self.dist = dist.copy()

    def loss(self, H):
        return np.zeros(H.shape)

    def npoints(self, H):
        return H.sum()

    def __call__(self, H):
        total_loss = np.sum(H * self.loss(H))
        if not self.renormalize:
            total_loss /= nonzero(self.npoints(H))
        return -total_loss


class SupervisedLikelihoodRatio(SimilarityMeasure):
    """
    Assume a joint intensity distribution model is given by self.dist
    """
    def loss(self, H):
        if not hasattr(self, 'L'):
            if self.dist is None:
                raise ValueError('SupervisedLikelihoodRatio: dist attribute cannot be None')
            if not self.dist.shape == H.shape:
                raise ValueError('SupervisedLikelihoodRatio: wrong shape for dist attribute')
            self.L = dist2loss(self.dist)
        return self.L


class MutualInformation(SimilarityMeasure):
    """
    Use the normalized joint histogram as a distribution model
    """
    def loss(self, H):
        return dist2loss(H / nonzero(self.npoints(H)))


class ParzenMutualInformation(SimilarityMeasure):
    """
    Use Parzen windowing to estimate the distribution model
    """
    def loss(self, H):
        if not hasattr(self, 'sigma'):
            self.sigma = SIGMA_FACTOR * np.array(H.shape)
        npts = nonzero(self.npoints(H))
        Hs = H / npts
        gaussian_filter(Hs, sigma=self.sigma, mode='constant', output=Hs)
        return dist2loss(Hs)


class DiscreteParzenMutualInformation(SimilarityMeasure):
    """
    Use Parzen windowing in the discrete case to estimate the
    distribution model
    """
    def loss(self, H):
        if not hasattr(self, 'sigma'):
            self.sigma = SIGMA_FACTOR * np.array(H.shape)
        Hs = gaussian_filter(H, sigma=self.sigma, mode='constant')
        Hs /= nonzero(Hs.sum())
        return dist2loss(Hs)


class NormalizedMutualInformation(SimilarityMeasure):
    """
    NMI = 2*(1 - H(I,J)/[H(I)+H(J)])
        = 2*MI/[H(I)+H(J)])
    """
    def __call__(self, H):
        H = H / nonzero(self.npoints(H))
        hI = H.sum(0)
        hJ = H.sum(1)
        entIJ = -np.sum(H * np.log(nonzero(H)))
        entI = -np.sum(hI * np.log(nonzero(hI)))
        entJ = -np.sum(hJ * np.log(nonzero(hJ)))
        return 2 * (1 - entIJ / nonzero(entI + entJ))


class CorrelationCoefficient(SimilarityMeasure):
    """
    Use a bivariate Gaussian as a distribution model
    """
    def loss(self, H):
        rho2 = self(H)
        I = (self.I - self.mI) / np.sqrt(nonzero(self.vI))
        J = (self.J - self.mJ) / np.sqrt(nonzero(self.vJ))
        L = rho2 * I ** 2 + rho2 * J ** 2 - 2 * self.rho * I * J
        tmp = nonzero(1. - rho2)
        L *= .5 / tmp
        L += .5 * np.log(tmp)
        return L

    def __call__(self, H):
        npts = nonzero(self.npoints(H))
        mI = np.sum(H * self.I) / npts
        mJ = np.sum(H * self.J) / npts
        vI = np.sum(H * (self.I) ** 2) / npts - mI ** 2
        vJ = np.sum(H * (self.J) ** 2) / npts - mJ ** 2
        cIJ = np.sum(H * self.J * self.I) / npts - mI * mJ
        rho2 = (cIJ / nonzero(np.sqrt(vI * vJ))) ** 2
        if self.renormalize:
            rho2 = correlation2loglikelihood(rho2, npts)
        return rho2


class CorrelationRatio(SimilarityMeasure):
    """
    Use a nonlinear regression model with Gaussian errors as a
    distribution model
    """
    def __call__(self, H):
        npts_J = np.sum(H, 1)
        tmp = nonzero(npts_J)
        mI_J = np.sum(H * self.I, 1) / tmp
        vI_J = np.sum(H * (self.I) ** 2, 1) / tmp - mI_J ** 2
        npts = np.sum(npts_J)
        tmp = nonzero(npts)
        hI = np.sum(H, 0)
        hJ = np.sum(H, 1)
        mI = np.sum(hI * self.I[0, :]) / tmp
        vI = np.sum(hI * self.I[0, :] ** 2) / tmp - mI ** 2
        mean_vI_J = np.sum(hJ * vI_J) / tmp
        eta2 = 1. - mean_vI_J / nonzero(vI)
        if self.renormalize:
            eta2 = correlation2loglikelihood(eta2, npts)
        return eta2


class CorrelationRatioL1(SimilarityMeasure):
    """
    Use a nonlinear regression model with Laplace distributed errors
    as a distribution model
    """
    def __call__(self, H):
        moments = np.array([_L1_moments(H[j, :]) for j in range(H.shape[0])])
        npts_J, mI_J, sI_J = moments[:, 0], moments[:, 1], moments[:, 2]
        hI = np.sum(H, 0)
        hJ = np.sum(H, 1)
        npts, mI, sI = _L1_moments(hI)
        mean_sI_J = np.sum(hJ * sI_J) / nonzero(npts)
        eta2 = 1. - mean_sI_J / nonzero(sI)
        if self.renormalize:
            eta2 = correlation2loglikelihood(eta2, npts)
        return eta2


similarity_measures = {
    'slr': SupervisedLikelihoodRatio,
    'mi': MutualInformation,
    'nmi': NormalizedMutualInformation,
    'pmi': ParzenMutualInformation,
    'dpmi': DiscreteParzenMutualInformation,
    'cc': CorrelationCoefficient,
    'cr': CorrelationRatio,
    'crl1': CorrelationRatioL1}



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
General utilities for code support.

These are modules that we (broadly-speaking) wrote; packages that other people
wrote, that we ship, go in the nipy.externals tree.
"""
from __future__ import absolute_import

from nibabel.data import make_datasource, DataError, datasource_or_bomber

# Module level datasource instances for convenience
from ..info import DATA_PKGS
templates = datasource_or_bomber(DATA_PKGS['nipy-templates'])
example_data = datasource_or_bomber(DATA_PKGS['nipy-data'])

try:
    example_data.get_filename()
except DataError:
    HAVE_EXAMPLE_DATA = False
else:
    HAVE_EXAMPLE_DATA = True

try:
    templates.get_filename()
except DataError:
    HAVE_TEMPLATES = False
else:
    HAVE_TEMPLATES = True

from .utilities import is_iterable, is_numlike, seq_prod

from nipy.testing import Tester
test = Tester().test
bench = Tester().bench


class VisibleDeprecationWarning(UserWarning):
    """Visible deprecation warning.

    Python does not show any DeprecationWarning by default.  Sometimes we do
    want to show a deprecation warning, when the deprecation is urgent, or the
    usage is probably a bug.
    """


class _NoValue:
    """Special keyword value.

    This class may be used as the default value assigned to a deprecated
    keyword in order to check if it has been given a user defined value.
    """



''' Tests for Euler angles '''
from __future__ import absolute_import

import math
import numpy as np
from numpy import pi

from .. import quaternions as tq
from .. import taitbryan as ttb

from nose.tools import assert_true, assert_false, assert_equal

from numpy.testing import assert_array_equal, assert_array_almost_equal

from .samples import euler_tuples

FLOAT_EPS = np.finfo(np.float).eps


def x_only(x):
    cosx = np.cos(x)
    sinx = np.sin(x)
    return np.array(
        [[1, 0, 0],
         [0, cosx, -sinx],
         [0, sinx, cosx]])

                 
def y_only(y):
    cosy = np.cos(y)
    siny = np.sin(y)
    return np.array(
        [[cosy, 0, siny],
         [0, 1, 0],
         [-siny, 0, cosy]])


def z_only(z):
    cosz = np.cos(z)
    sinz = np.sin(z)
    return np.array(
                [[cosz, -sinz, 0],
                 [sinz, cosz, 0],
                 [0, 0, 1]])


def sympy_euler(z, y, x):
    # The whole matrix formula for z,y,x rotations from Sympy
    cos = math.cos
    sin = math.sin
    # the following copy / pasted from Sympy - see derivations subdirectory
    return [
        [                       cos(y)*cos(z),                       -cos(y)*sin(z),         sin(y)],
        [cos(x)*sin(z) + cos(z)*sin(x)*sin(y), cos(x)*cos(z) - sin(x)*sin(y)*sin(z), -cos(y)*sin(x)],
        [sin(x)*sin(z) - cos(x)*cos(z)*sin(y), cos(z)*sin(x) + cos(x)*sin(y)*sin(z),  cos(x)*cos(y)]
        ]


def is_valid_rotation(M):
    if not np.allclose(np.linalg.det(M), 1):
        return False
    return np.allclose(np.eye(3), np.dot(M, M.T))


def test_basic_euler():
    # some example rotations, in radians
    zr = 0.05
    yr = -0.4
    xr = 0.2
    # Rotation matrix composing the three rotations
    M = ttb.euler2mat(zr, yr, xr)
    # Corresponding individual rotation matrices
    M1 = ttb.euler2mat(zr)
    M2 = ttb.euler2mat(0, yr)
    M3 = ttb.euler2mat(0, 0, xr)
    # which are all valid rotation matrices
    yield assert_true, is_valid_rotation(M)
    yield assert_true, is_valid_rotation(M1)
    yield assert_true, is_valid_rotation(M2)
    yield assert_true, is_valid_rotation(M3)
    # Full matrix is composition of three individual matrices
    yield assert_true, np.allclose(M, np.dot(M3, np.dot(M2, M1)))
    # Rotations can be specified with named args, default 0
    yield assert_true, np.all(ttb.euler2mat(zr) == ttb.euler2mat(z=zr))
    yield assert_true, np.all(ttb.euler2mat(0, yr) == ttb.euler2mat(y=yr))
    yield assert_true, np.all(ttb.euler2mat(0, 0, xr) == ttb.euler2mat(x=xr))
    # Applying an opposite rotation same as inverse (the inverse is
    # the same as the transpose, but just for clarity)
    yield assert_true, np.allclose(ttb.euler2mat(x=-xr),
                       np.linalg.inv(ttb.euler2mat(x=xr)))

        
def test_euler_mat():
    M = ttb.euler2mat()
    yield assert_array_equal, M, np.eye(3)
    for x, y, z in euler_tuples:
        M1 = ttb.euler2mat(z, y, x)
        M2 = sympy_euler(z, y, x)
        yield assert_array_almost_equal, M1, M2
        M3 = np.dot(x_only(x), np.dot(y_only(y), z_only(z)))
        yield assert_array_almost_equal, M1, M3
        zp, yp, xp = ttb.mat2euler(M1)
        # The parameters may not be the same as input, but they give the
        # same rotation matrix
        M4 = ttb.euler2mat(zp, yp, xp)
        yield assert_array_almost_equal, M1, M4


def sympy_euler2quat(z=0, y=0, x=0):
    # direct formula for z,y,x quaternion rotations using sympy
    # see derivations subfolder
    cos = math.cos
    sin = math.sin
    # the following copy / pasted from Sympy output
    return (cos(0.5*x)*cos(0.5*y)*cos(0.5*z) - sin(0.5*x)*sin(0.5*y)*sin(0.5*z),
            cos(0.5*x)*sin(0.5*y)*sin(0.5*z) + cos(0.5*y)*cos(0.5*z)*sin(0.5*x),
            cos(0.5*x)*cos(0.5*z)*sin(0.5*y) - cos(0.5*y)*sin(0.5*x)*sin(0.5*z),
            cos(0.5*x)*cos(0.5*y)*sin(0.5*z) + cos(0.5*z)*sin(0.5*x)*sin(0.5*y))


def crude_mat2euler(M):
    ''' The simplest possible - ignoring atan2 instability '''
    r11, r12, r13, r21, r22, r23, r31, r32, r33 = M.flat
    return math.atan2(-r12, r11), math.asin(r13), math.atan2(-r23, r33)


def test_euler_instability():
    # Test for numerical errors in mat2euler
    # problems arise for cos(y) near 0
    po2 = pi / 2
    zyx = po2, po2, po2
    M = ttb.euler2mat(*zyx)
    # Round trip
    M_back = ttb.euler2mat(*ttb.mat2euler(M))
    yield assert_true, np.allclose(M, M_back)
    # disturb matrix slightly
    M_e = M - FLOAT_EPS
    # round trip to test - OK
    M_e_back = ttb.euler2mat(*ttb.mat2euler(M_e))
    yield assert_true, np.allclose(M_e, M_e_back)
    # not so with crude routine
    M_e_back = ttb.euler2mat(*crude_mat2euler(M_e))
    yield assert_false, np.allclose(M_e, M_e_back)


def test_quats():
    for x, y, z in euler_tuples:
        M1 = ttb.euler2mat(z, y, x)
        quatM = tq.mat2quat(M1)
        quat = ttb.euler2quat(z, y, x)
        yield tq.nearly_equivalent, quatM, quat
        quatS = sympy_euler2quat(z, y, x)
        yield tq.nearly_equivalent, quat, quatS
        zp, yp, xp = ttb.quat2euler(quat)
        # The parameters may not be the same as input, but they give the
        # same rotation matrix
        M2 = ttb.euler2mat(zp, yp, xp)
        yield assert_array_almost_equal, M1, M2
        



""" Implementation of diagnostic command line tools

Tools are:

* nipy_diagnose
* nipy_tsdiffana

This module has the logic for each command.

The command script files deal with argument parsing and any custom imports.
The implementation here accepts the ``args`` object from ``argparse`` and does
the work.
"""
from __future__ import absolute_import
from os.path import split as psplit, join as pjoin

import numpy as np

from nibabel import AnalyzeHeader
from nibabel.filename_parser import splitext_addext

import nipy

from .tsdiffplot import plot_tsdiffs
from .timediff import time_slice_diffs_image
from .screens import screen, write_screen_res


def parse_fname_axes(img_fname, time_axis, slice_axis):
    """ Load `img_fname`, check `time_axis`, `slice_axis` or use default

    Parameters
    ----------
    img_fname : str
       filename of image on which to do diagnostics
    time_axis : None or str or int, optional
        Axis indexing time-points. None is default, will be replaced by a value
        of 't'. If `time_axis` is an integer, gives the index of the input
        (domain) axis of `img`. If `time_axis` is a str, can be an input
        (domain) name, or an output (range) name, that maps to an input
        (domain) name.
    slice_axis : None or str or int, optional
        Axis indexing MRI slices. If `slice_axis` is an integer, gives the
        index of the input (domain) axis of `img`. If `slice_axis` is a str,
        can be an input (domain) name, or an output (range) name, that maps to
        an input (domain) name.  If None (the default) then 1) try the name
        'slice' to select the axis - if this fails, and `fname` refers to an
        Analyze type image (such as Nifti), then 2) default to the third image
        axis, otherwise 3) raise a ValueError

    Returns
    -------
    img : ``Image`` instance
        Image as loaded from `img_fname`
    time_axis : int or str
        Time axis, possibly filled with default
    slice_axis : int or str
        Slice axis, possibly filled with default
    """
    # Check whether this is an Analyze-type image
    img = nipy.load_image(img_fname)
    # Check for axes
    if time_axis is not None:
        # Try converting to an integer in case that was what was passed
        try:
            time_axis = int(time_axis)
        except ValueError:
            # Maybe a string
            pass
    else: # was None
        time_axis = 't'
    if slice_axis is not None:
        # Try converting to an integer in case that was what was passed
        try:
            slice_axis = int(slice_axis)
        except ValueError:
            # Maybe a string
            pass
    else: # slice axis was None - search for default
        input_names = img.coordmap.function_domain.coord_names
        is_analyze = ('header' in img.metadata and
                      isinstance(img.metadata['header'], AnalyzeHeader))
        if 'slice' in input_names:
            slice_axis = 'slice'
        elif is_analyze and img.ndim == 4:
            slice_axis = 2
        else:
            raise ValueError('No slice axis specified, not analyze type '
                             'image; refusing to guess')
    return img, time_axis, slice_axis


def tsdiffana(args):
    """ Generate tsdiffana plots from command line params `args`

    Parameters
    ----------
    args : object
        object with attributes

        * filename : str - 4D image filename
        * out_file : str - graphics file to write to instead of leaving
          graphics on screen
        * time_axis : str - name or number of time axis in `filename`
        * slice_axis : str - name or number of slice axis in `filename`
        * write_results : bool - if True, write images and plots to files
        * out_path : None or str - path to which to write results
        * out_fname_label : None or filename - suffix of output results files

    Returns
    -------
    axes : Matplotlib axes
       Axes on which we have done the plots.
    """
    if args.out_file is not None and args.write_results:
        raise ValueError("Cannot have OUT_FILE and WRITE_RESULTS options "
                         "together")
    img, time_axis, slice_axis = parse_fname_axes(args.filename,
                                                  args.time_axis,
                                                  args.slice_axis)
    results = time_slice_diffs_image(img, time_axis, slice_axis)
    axes = plot_tsdiffs(results)
    if args.out_file is None and not args.write_results:
        # interactive mode
        return axes
    if args.out_file is not None:
        # plot only mode
        axes[0].figure.savefig(args.out_file)
        return axes
    # plot and images mode
    froot, ext, addext = splitext_addext(args.filename)
    fpath, fbase = psplit(froot)
    fpath = fpath if args.out_path is None else args.out_path
    fbase = fbase if args.out_fname_label is None else args.out_fname_label
    axes[0].figure.savefig(pjoin(fpath, 'tsdiff_' + fbase + '.png'))
    # Save image volumes
    for key, prefix in (('slice_diff2_max_vol', 'dv2_max_'),
                        ('diff2_mean_vol', 'dv2_mean_')):
        fname = pjoin(fpath, prefix + fbase + ext + addext)
        nipy.save_image(results[key], fname)
    # Save time courses into npz
    np.savez(pjoin(fpath, 'tsdiff_' + fbase + '.npz'),
             volume_means=results['volume_means'],
             slice_mean_diff2=results['slice_mean_diff2'],
            )
    return axes


def diagnose(args):
    """ Calculate, write results from diagnostic screen

    Parameters
    ----------
    args : object
        object with attributes:

        * filename : str - 4D image filename
        * time_axis : str - name or number of time axis in `filename`
        * slice_axis : str - name or number of slice axis in `filename`
        * out_path : None or str - path to which to write results
        * out_fname_label : None or filename - suffix of output results files
        * ncomponents : int - number of PCA components to write images for

    Returns
    -------
    res : dict
        Results of running :func:`screen` on `filename`
    """
    img, time_axis, slice_axis = parse_fname_axes(args.filename,
                                                  args.time_axis,
                                                  args.slice_axis)
    res = screen(img, args.ncomponents, time_axis, slice_axis)
    froot, ext, addext = splitext_addext(args.filename)
    fpath, fbase = psplit(froot)
    fpath = fpath if args.out_path is None else args.out_path
    fbase = fbase if args.out_fname_label is None else args.out_fname_label
    write_screen_res(res, fpath, fbase, ext + addext)
    return res



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
""" Testing diagnostic screen
"""
from __future__ import absolute_import

import os
from os.path import join as pjoin

from warnings import catch_warnings, simplefilter

import numpy as np

import nipy as ni
from nipy.core.api import rollimg
from ..screens import screen, write_screen_res
from ..timediff import time_slice_diffs
from ...utils.pca import pca
from ...utils.tests.test_pca import res2pos1

from nibabel.tmpdirs import InTemporaryDirectory

from nose.tools import (assert_true, assert_false, assert_equal, assert_raises)

from numpy.testing import (assert_array_equal, assert_array_almost_equal,
                           assert_almost_equal, decorators)

from nipy.testing import funcfile
from nipy.testing.decorators import needs_mpl_agg


def _check_pca(res, pca_res):
    # Standardize output vector signs
    screen_pca_res = res2pos1(res['pca_res'])
    for key in pca_res:
        assert_almost_equal(pca_res[key], screen_pca_res[key])


def _check_ts(res, data, time_axis, slice_axis):
    ts_res = time_slice_diffs(data, time_axis, slice_axis)
    for key in ts_res:
        assert_array_equal(ts_res[key], res['ts_res'][key])


def test_screen():
    img = ni.load_image(funcfile)
    # rename third axis to slice to match default of screen
    # This avoids warnings about future change in default; see the tests for
    # slice axis below
    img = img.renamed_axes(k='slice')
    res = screen(img)
    assert_equal(res['mean'].ndim, 3)
    assert_equal(res['pca'].ndim, 4)
    assert_equal(sorted(res.keys()),
                 ['max', 'mean', 'min',
                  'pca', 'pca_res',
                  'std', 'ts_res'])
    data = img.get_data()
    # Check summary images
    assert_array_equal(np.max(data, axis=-1), res['max'].get_data())
    assert_array_equal(np.mean(data, axis=-1), res['mean'].get_data())
    assert_array_equal(np.min(data, axis=-1), res['min'].get_data())
    assert_array_equal(np.std(data, axis=-1), res['std'].get_data())
    pca_res = pca(data, axis=-1, standardize=False, ncomp=10)
    # On windows, there seems to be some randomness in the PCA output vector
    # signs; this routine sets the basis vectors to have first value positive,
    # and therefore standardizes the signs
    pca_res = res2pos1(pca_res)
    _check_pca(res, pca_res)
    _check_ts(res, data, 3, 2)
    # Test that screens accepts and uses time axis
    data_mean = data.mean(axis=-1)
    res = screen(img, time_axis='t')
    assert_array_equal(data_mean, res['mean'].get_data())
    _check_pca(res, pca_res)
    _check_ts(res, data, 3, 2)
    res = screen(img, time_axis=-1)
    assert_array_equal(data_mean, res['mean'].get_data())
    _check_pca(res, pca_res)
    _check_ts(res, data, 3, 2)
    t0_img = rollimg(img, 't')
    t0_data = np.rollaxis(data, -1)
    res = screen(t0_img, time_axis='t')
    t0_pca_res = pca(t0_data, axis=0, standardize=False, ncomp=10)
    t0_pca_res = res2pos1(t0_pca_res)
    assert_array_equal(data_mean, res['mean'].get_data())
    _check_pca(res, t0_pca_res)
    _check_ts(res, t0_data, 0, 3)
    res = screen(t0_img, time_axis=0)
    assert_array_equal(data_mean, res['mean'].get_data())
    _check_pca(res, t0_pca_res)
    _check_ts(res, t0_data, 0, 3)
    # Check screens uses slice axis
    s0_img = rollimg(img, 2, 0)
    s0_data = np.rollaxis(data, 2, 0)
    res = screen(s0_img, slice_axis=0)
    _check_ts(res, s0_data, 3, 0)
    # And defaults to named slice axis
    # First re-show that when we don't specify, we get the default
    res = screen(img)
    _check_ts(res, data, 3, 2)
    assert_raises(AssertionError, _check_ts, res, data, 3, 0)
    # Then specify, get non-default
    slicey_img = img.renamed_axes(slice='k', i='slice')
    res = screen(slicey_img)
    _check_ts(res, data, 3, 0)
    assert_raises(AssertionError, _check_ts, res, data, 3, 2)


def pca_pos(data4d):
    """ Flips signs equal over volume for PCA

    Needed because Windows appears to generate random signs for PCA components
    across PCA runs on the same data.
    """
    signs = np.sign(data4d[0, 0, 0, :])
    return data4d * signs


def test_screen_slice_axis():
    img = ni.load_image(funcfile)
    # Default screen raises a FutureWarning because of the default slice_axis
    exp_res = screen(img, slice_axis='k')
    with catch_warnings():
        simplefilter('error')
        assert_raises(FutureWarning, screen, img)
        assert_raises(FutureWarning, screen, img, slice_axis=None)
        explicit_img = img.renamed_axes(k='slice')
        # Now the analysis works without warning
        res = screen(explicit_img)
        # And is the expected analysis
        # Very oddly on scipy 0.9 32 bit - at least - results differ between
        # runs, so we need assert_almost_equal
        assert_almost_equal(pca_pos(res['pca'].get_data()),
                            pca_pos(exp_res['pca'].get_data()))
        assert_array_equal(res['ts_res']['slice_mean_diff2'],
                           exp_res['ts_res']['slice_mean_diff2'])
        # Turn off warnings, also get expected analysis
        simplefilter('ignore')
        res = screen(img)
        assert_array_equal(res['ts_res']['slice_mean_diff2'],
                           exp_res['ts_res']['slice_mean_diff2'])


@needs_mpl_agg
def test_write_screen_res():
    img = ni.load_image(funcfile)
    with InTemporaryDirectory():
        res = screen(img)
        os.mkdir('myresults')
        write_screen_res(res, 'myresults', 'myana')
        pca_img = ni.load_image(pjoin('myresults', 'pca_myana.nii'))
        assert_equal(pca_img.shape, img.shape[:-1] + (10,))
        # Make sure we get the same output image even from rolled image
        # Do fancy roll to put time axis first, and slice axis last. This does
        # a stress test on the axis ordering, but also makes sure that we are
        # getting the number of components from the right place.  If we were
        # getting the number of components from the length of the last axis,
        # instead of the length of the 't' axis in the returned pca image, this
        # would be wrong (=21) which would also be more than the number of
        # basis vectors (19) so raise an error
        rimg = img.reordered_axes([3, 2, 0, 1])
        os.mkdir('rmyresults')
        rres = screen(rimg)
        write_screen_res(rres, 'rmyresults', 'myana')
        rpca_img = ni.load_image(pjoin('rmyresults', 'pca_myana.nii'))
        assert_equal(rpca_img.shape, img.shape[:-1] + (10,))
        del pca_img, rpca_img



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
This module is the interface to the bayesian_structural_analysis (bsa) module
It handles the images provided as input and produces result images.
"""
from __future__ import absolute_import

import numpy as np
import os.path as op
from nibabel import load, save, Nifti1Image

from ..mask import intersect_masks
from .bayesian_structural_analysis import compute_landmarks
from .discrete_domain import domain_from_image
from ...io.nibcompat import get_header, get_affine

from ...externals.six import string_types

def make_bsa_image(
    mask_images, stat_images, threshold=3., smin=0, sigma=5.,
    prevalence_threshold=0, prevalence_pval=0.5, write_dir=None,
    algorithm='density', contrast_id='default'):
    """ Main function for  performing bsa on a set of images.
    It creates the some output images in the given directory

    Parameters
    ----------
    mask_images: list of str,
                 image paths that yield mask images, one for each subject
    stat_images: list of str,
                 image paths of the activation images, one for each subject
    threshold: float, optional,
               threshold used to ignore all the image data that is below
    smin: float, optional,
          minimal size (in voxels) of the extracted blobs
          smaller blobs are merged into larger ones
    sigma: float, optional,
           variance of the spatial model, i.e. cross-subject uncertainty
    prevalence_threshold: float, optional
                          threshold on the representativity measure
    prevalence_pval: float, optional,
                     p-value of the representativity test:
             test = p(representativity>prevalence_threshold) > prevalence_pval
    write_dir: string, optional,
               if not None, output directory
    method: {'density', 'co-occurrence'}, optional,
            Inference method used in the landmark definition
    contrast_id: string, optional,
                 identifier of the contrast

    Returns
    -------
    landmarks: nipy.labs.spatial_models.structural_bfls.landmark_regions
         instance that describes the structures found at the group level
         None is returned if nothing has been found significant
         at the group level
    hrois : list of nipy.labs.spatial_models.hroi.Nroi instances,
       (one per subject), describe the individual counterpart of landmarks
    """
    n_subjects = len(stat_images)

    # Read the referential information
    nim = load(stat_images[0])
    ref_dim = nim.shape[:3]
    affine = get_affine(nim)

    # Read the masks and compute the "intersection"
    # mask = np.reshape(intersect_masks(mask_images), ref_dim).astype('u8')
    if isinstance(mask_images, string_types):
        mask = load(mask_images).get_data()
    elif isinstance(mask_images, Nifti1Image):
        mask = mask_images.get_data()
    else:
        # mask_images should be a list of strings or images
        mask = intersect_masks(mask_images).astype('u8')

    # encode it as a domain
    domain = domain_from_image(Nifti1Image(mask, affine), nn=18)
    n_voxels = domain.size

    # read the functional images
    stats = []
    for stat_image in stat_images:
        beta = np.reshape(load(stat_image).get_data(), ref_dim)
        stats.append(beta[mask > 0])
    stats = np.array(stats).T

    # launch the method
    crmap = - np.ones(n_voxels).astype(np.int16)
    density = np.zeros(n_voxels)
    landmarks = None
    hrois = [None for _ in range(n_subjects)]
    default_idx = 0
    prevalence = np.array([])

    landmarks, hrois = compute_landmarks(
        domain, stats, sigma, prevalence_pval, prevalence_threshold,
        threshold, smin, algorithm=algorithm)

    if landmarks is not None:
        crmap = landmarks.map_label(domain.coord, 0.95, sigma).astype(np.int16)
        density = landmarks.kernel_density(
            k=None, coord=domain.coord, sigma=sigma)
        default_idx = landmarks.k + 2
        prevalence = landmarks.roi_prevalence()

    if write_dir == False:
        return landmarks, hrois

    # Write the results as images

    # the spatial density image
    density_map = np.zeros(ref_dim)
    density_map[mask > 0] = density
    wim = Nifti1Image(density_map, affine)
    get_header(wim)['descrip'] = ('group-level spatial density '
                                  'of active regions')
    dens_path = op.join(write_dir, "density_%s.nii" % contrast_id)
    save(wim, dens_path)

    # write a 3D image for group-level labels
    labels = - 2 * np.ones(ref_dim)
    labels[mask > 0] = crmap
    wim = Nifti1Image(labels.astype('int16'), affine)
    get_header(wim)['descrip'] = 'group Level labels from bsa procedure'
    save(wim, op.join(write_dir, "CR_%s.nii" % contrast_id))

    # write a prevalence image
    prev_ = np.zeros(crmap.size).astype(np.float)
    prev_[crmap > -1] = prevalence[crmap[crmap > -1]]
    prevalence_map = - np.ones(ref_dim)
    prevalence_map[mask > 0] = prev_
    wim = Nifti1Image(prevalence_map, affine)
    get_header(wim)['descrip'] = 'Weighted prevalence image'
    save(wim, op.join(write_dir, "prevalence_%s.nii" % contrast_id))

    # write a 4d images with all subjects results
    wdim = (ref_dim[0], ref_dim[1], ref_dim[2], n_subjects)
    labels = - 2 * np.ones(wdim, 'int16')
    for subject in range(n_subjects):
        labels[mask > 0, subject] = - 1
        if hrois[subject].k > 0:
            nls = hrois[subject].get_roi_feature('label')
            nls[nls == - 1] = default_idx
            lab = hrois[subject].label
            lab[lab > - 1] = nls[lab[lab > - 1]]
            labels[mask > 0, subject] = lab
    wim = Nifti1Image(labels, affine)
    get_header(wim)['descrip'] = 'Individual labels from bsa procedure'
    save(wim, op.join(write_dir, "AR_%s.nii" % contrast_id))
    return landmarks, hrois



"""
Utilities to test whether a variable is of, or convertible to, a particular type
"""
import numpy as np


def _check_type(x, t):
    try:
        y = t(x)
        return True
    except:
        return False


def check_type(x, t, accept_none=False):
    """
    Checks whether a variable is convertible to a certain type.
    A ValueError is raised if test fails.

    Parameters
    ----------
    x : object
        Input argument to be checked.
    t : type
        Target type.
    accept_none : bool
        If True, skip errors if `x` is None.
    """	
    if accept_none:
        if x is None:
            return
    if not _check_type(x, t):
        raise ValueError('Argument should be convertible to %s' % t)


def check_type_and_shape(x, t, s, accept_none=False):
    """
    Checks whether a sequence is convertible to a numpy ndarray with
    given shape, and if the elements are convertible to a certain type.
    A ValueError is raised if test fails.

    Parameters
    ----------
    x : sequence
        Input sequence to be checked.
    t : type
        Target element-wise type.
    s : sequence of ints
        Target shape.
    accept_none : bool
        If True, skip errors if `x` is None.
    """
    if accept_none:
        if x is None:
            return
    try:
        shape = (int(s), )
    except:
        shape = tuple(s)
    try:
        y = np.asarray(x)
        ok_type = _check_type(y[0], t)
        ok_shape = (y.shape == shape)
    except:
        raise ValueError('Argument should be convertible to ndarray')
    if not ok_type:
        raise ValueError('Array values should be convertible to %s' % t)
    if not ok_shape:
        raise ValueError('Array shape should be equivalent to %s' % shape)




from __future__ import absolute_import
import numpy as np
from nose.tools import raises
from numpy.testing import (
    assert_almost_equal, assert_equal, assert_array_equal, assert_warns)
import warnings

from ..hemodynamic_models import (
    spm_hrf, spm_time_derivative, spm_dispersion_derivative,
    _resample_regressor, _orthogonalize, _sample_condition,
    _regressor_names, _hrf_kernel, glover_hrf, 
    glover_time_derivative, compute_regressor)


def test_spm_hrf():
    """ test that the spm_hrf is correctly normalized and has correct length
    """
    h = spm_hrf(2.0)
    assert_almost_equal(h.sum(), 1)
    assert_equal(len(h), 256)

def test_spm_hrf_derivative():
    """ test that the spm_hrf is correctly normalized and has correct length
    """
    h = spm_time_derivative(2.0)
    assert_almost_equal(h.sum(), 0)
    assert_equal(len(h), 256)
    h = spm_dispersion_derivative(2.0)
    assert_almost_equal(h.sum(), 0)
    assert_equal(len(h), 256)

def test_glover_hrf():
    """ test that the spm_hrf is correctly normalized and has correct length
    """
    h = glover_hrf(2.0)
    assert_almost_equal(h.sum(), 1)
    assert_equal(len(h), 256)

def test_glover_time_derivative():
    """ test that the spm_hrf is correctly normalized and has correct length
    """
    h = glover_time_derivative(2.0)
    assert_almost_equal(h.sum(), 0)
    assert_equal(len(h), 256)
    
def test_resample_regressor():
    """ test regressor resampling on a linear function
    """
    x = np.linspace(0, 1, 200)
    y = np.linspace(0, 1, 30)
    z = _resample_regressor(x, x, y)
    assert_almost_equal(z, y)

def test_resample_regressor_nl():
    """ test regressor resampling on a sine function
    """
    x = np.linspace(0, 10, 1000)
    y = np.linspace(0, 10, 30)
    z = _resample_regressor(np.cos(x), x, y)
    assert_almost_equal(z, np.cos(y), decimal=2)

def test_orthogonalize():
    """ test that the orthogonalization is OK 
    """
    X = np.random.randn(100, 5)
    X = _orthogonalize(X)
    K = np.dot(X.T, X)
    K -= np.diag(np.diag(K))
    assert_almost_equal((K ** 2).sum(), 0, 15)

def test_orthogonalize_trivial():
    """ test that the orthogonalization is OK 
    """
    X = np.random.randn(100)
    Y = X.copy()
    X = _orthogonalize(X)
    assert_array_equal(Y, X)

def test_sample_condition_1():
    """ Test that the experimental condition is correctly sampled
    """
    condition = ([1, 20, 36.5], [0, 0, 0], [1, 1, 1])
    frametimes = np.linspace(0, 49, 50)
    reg, rf = _sample_condition(condition, frametimes, oversampling=1, 
                               min_onset=0)
    assert_equal(reg.sum(), 3)
    assert_equal(reg[1], 1)
    assert_equal(reg[20], 1)
    assert_equal(reg[37], 1)

    reg, rf = _sample_condition(condition, frametimes, oversampling=1)
    assert_equal(reg.sum(), 3)
    assert_equal(reg[25], 1)
    assert_equal(reg[44], 1)
    assert_equal(reg[61], 1)


def test_sample_condition_2():
    """ Test the experimental condition sampling -- onset = 0
    """
    condition = ([0, 20, 36.5], [2, 2, 2], [1, 1, 1])
    frametimes = np.linspace(0, 49, 50)
    reg, rf = _sample_condition(condition, frametimes, oversampling=1,
                               min_onset=- 10)
    assert_equal(reg.sum(), 6)
    assert_equal(reg[10], 1)
    assert_equal(reg[48], 1)
    assert_equal(reg[31], 1)

def test_sample_condition_3():
    """ Test the experimental condition sampling -- oversampling=10
    """
    condition = ([1, 20, 36.5], [2, 2, 2], [1, 1, 1])
    frametimes = np.linspace(0, 49, 50)
    reg, rf = _sample_condition(condition, frametimes, oversampling=10,
                               min_onset=0)
    assert_almost_equal(reg.sum(), 60.)
    assert_equal(reg[10], 1)
    assert_equal(reg[380], 1)
    assert_equal(reg[210], 1)
    assert_equal(np.sum(reg > 0), 60)

def test_sample_condition_4():
    """ Test the experimental condition sampling -- negative amplitude
    """
    condition = ([1, 20, 36.5], [2, 2, 2], [1., -1., 5.])
    frametimes = np.linspace(0, 49, 50)
    reg, rf = _sample_condition(condition, frametimes, oversampling=1)
    assert_equal(reg.sum(),10)
    assert_equal(reg[25], 1.)
    assert_equal(reg[44], -1.)
    assert_equal(reg[61], 5.)

def test_sample_condition_5():
    """ Test the experimental condition sampling -- negative onset
    """
    condition = ([-10, 0, 36.5], [2, 2, 2], [1., -1., 5.])
    frametimes = np.linspace(0, 49, 50)
    reg, rf = _sample_condition(condition, frametimes, oversampling=1)
    assert_equal(reg.sum(),10)
    assert_equal(reg[14], 1.)
    assert_equal(reg[24], -1.)
    assert_equal(reg[61], 5.)

def test_names():
    """ Test the regressor naming function
    """
    name = 'con'
    assert_equal(_regressor_names(name, 'spm'), ['con'])
    assert_equal(_regressor_names(name, 'spm_time'), ['con', 'con_derivative'])
    assert_equal(_regressor_names(name, 'spm_time_dispersion'),
        ['con', 'con_derivative', 'con_dispersion'])
    assert_equal(_regressor_names(name, 'canonical'), ['con'])
    assert_equal(_regressor_names(name, 'canonical with derivative'),
        ['con', 'con_derivative'])

def test_hkernel():
    """ test the hrf computation
    """
    tr = 2.0
    h = _hrf_kernel('spm', tr)
    assert_almost_equal(h[0], spm_hrf(tr))
    assert_equal(len(h), 1)
    h = _hrf_kernel('spm_time', tr)
    assert_almost_equal(h[1], spm_time_derivative(tr))
    assert_equal(len(h), 2)
    h = _hrf_kernel('spm_time_dispersion', tr)
    assert_almost_equal(h[2], spm_dispersion_derivative(tr))
    assert_equal(len(h), 3)
    h = _hrf_kernel('canonical', tr)
    assert_almost_equal(h[0], glover_hrf(tr))
    assert_equal(len(h), 1)
    h = _hrf_kernel('canonical with derivative', tr)
    assert_almost_equal(h[1], glover_time_derivative(tr))
    assert_almost_equal(h[0], glover_hrf(tr))
    assert_equal(len(h), 2)
    h = _hrf_kernel('fir', tr, fir_delays = np.arange(4))
    assert_equal(len(h), 4)
    for dh in h:
        assert_equal(dh.sum(), 16.)
    
def test_make_regressor_1():
    """ test the generated regressor
    """
    condition = ([1, 20, 36.5], [2, 2, 2], [1, 1, 1])
    frametimes = np.linspace(0, 69, 70)
    hrf_model = 'spm'
    reg, reg_names = compute_regressor(condition, hrf_model, frametimes)
    assert_almost_equal(reg.sum(), 6, 1)
    assert_equal(reg_names[0], 'cond')

def test_make_regressor_2():
    """ test the generated regressor
    """
    condition = ([1, 20, 36.5], [0, 0, 0], [1, 1, 1])
    frametimes = np.linspace(0, 69, 70)
    hrf_model = 'spm'
    reg, reg_names = compute_regressor(condition, hrf_model, frametimes)
    assert_almost_equal(reg.sum() * 16, 3, 1)
    assert_equal(reg_names[0], 'cond')


def test_make_regressor_3():
    """ test the generated regressor
    """
    condition = ([1, 20, 36.5], [0, 0, 0], [1, 1, 1])
    frametimes = np.linspace(0, 138, 70)
    hrf_model = 'fir'
    reg, reg_names = compute_regressor(condition, hrf_model, frametimes, 
                                       fir_delays=np.arange(4))
    assert_array_equal(np.unique(reg), np.array([0, 1]))
    assert_array_equal(np.sum(reg, 0), np.array([3, 3, 3, 3]))
    assert_equal(len(reg_names), 4)

def test_design_warnings():
    """ test that warnings are correctly raised upon weird design specification
    """
    condition = ([-25, 20, 36.5], [0, 0, 0], [1, 1, 1])
    frametimes = np.linspace(0, 69, 70)
    hrf_model = 'spm'
    with warnings.catch_warnings(record=True):
        warnings.simplefilter("always")
        assert_warns(UserWarning, compute_regressor, condition, hrf_model, 
                     frametimes)
    condition = ([-25, -25, 36.5], [0, 0, 0], [1, 1, 1])
    with warnings.catch_warnings(record=True):
        warnings.simplefilter("always")
        assert_warns(UserWarning, compute_regressor, condition, hrf_model, 
                     frametimes)

if __name__ == "__main__":
    import nose
    nose.run(argv=['', __file__])



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
""" Testing tsdiffana

"""
from __future__ import absolute_import

from os.path import dirname, join as pjoin

import numpy as np

import scipy.io as sio

from ....core.api import rollimg
from ....core.reference.coordinate_map import AxisError

from .. import timediff as tsd

from nose.tools import (assert_true, assert_false, assert_equal,
                        assert_not_equal, assert_raises)

from numpy.testing import assert_array_equal, assert_array_almost_equal

from nipy import load_image
from nipy.testing import funcfile

TEST_DATA_PATH = pjoin(dirname(__file__), 'data')


def test_time_slice_diffs():
    n_tps = 10
    n_slices = 4
    slice_shape = (2,3)
    slice_size = np.prod(slice_shape)
    vol_shape = slice_shape + (n_slices,)
    vol_size = np.prod(vol_shape)
    ts = np.random.normal(size=vol_shape + (n_tps,)) * 100 + 10
    expected = {}
    expected['volume_means'] = ts.reshape((vol_size, -1)).mean(0)
    # difference over time ^2
    diffs2 = np.diff(ts, axis=-1)**2
    expected['volume_mean_diff2'] = np.mean(
        diffs2.reshape((vol_size, -1)), 0)
    expected['slice_mean_diff2'] = np.zeros((n_tps-1, n_slices))
    for s in range(n_slices):
        v = diffs2[:,:,s,:].reshape((slice_size, -1))
        expected['slice_mean_diff2'][:,s] = np.mean(v, 0)
    expected['diff2_mean_vol'] = np.mean(diffs2, -1)
    max_diff_is = np.argmax(expected['slice_mean_diff2'], 0)
    sdmv = np.empty(vol_shape)
    for si, dti in enumerate(max_diff_is):
        sdmv[:,:,si] = diffs2[:,:,si,dti]
    expected['slice_diff2_max_vol'] = sdmv
    results = tsd.time_slice_diffs(ts)
    for key in expected:
        assert_array_almost_equal(results[key], expected[key])
    # tranposes, reset axes, get the same result
    results = tsd.time_slice_diffs(ts.T, 0, 1)
    results['diff2_mean_vol'] = results['diff2_mean_vol'].T
    results['slice_diff2_max_vol'] = results['slice_diff2_max_vol'].T
    for key in expected:
        assert_array_almost_equal(results[key], expected[key])
    ts_t = ts.transpose((1, 3, 0, 2))
    results = tsd.time_slice_diffs(ts_t, 1, -1)
    results['diff2_mean_vol'] = results['diff2_mean_vol'].transpose(
        ((1,0,2)))
    results['slice_diff2_max_vol'] = results['slice_diff2_max_vol'].transpose(
        ((1,0,2)))
    for key in expected:
        assert_array_almost_equal(results[key], expected[key])


def test_time_slice_axes():
    # Test time and slice axes work as expected
    fimg = load_image(funcfile)
    # Put into array
    data = fimg.get_data()
    orig_results = tsd.time_slice_diffs(data)
    t0_data = np.rollaxis(data, 3)
    t0_results = tsd.time_slice_diffs(t0_data, 0)
    for key in ('volume_means', 'slice_mean_diff2'):
        assert_array_almost_equal(orig_results[key], t0_results[key])
    s0_data = np.rollaxis(data, 2)
    s0_results = tsd.time_slice_diffs(s0_data, slice_axis=0)
    for key in ('volume_means', 'slice_mean_diff2'):
        assert_array_almost_equal(orig_results[key], s0_results[key])
    # Incorrect slice axis
    bad_s0_results = tsd.time_slice_diffs(s0_data)
    assert_not_equal(orig_results['slice_mean_diff2'].shape,
                     bad_s0_results['slice_mean_diff2'].shape)
    # Slice axis equal to time axis - ValueError
    assert_raises(ValueError, tsd.time_slice_diffs, data, -1, -1)
    assert_raises(ValueError, tsd.time_slice_diffs, data, -1, 3)
    assert_raises(ValueError, tsd.time_slice_diffs, data, 1, 1)
    assert_raises(ValueError, tsd.time_slice_diffs, data, 1, -3)


def test_against_matlab_results():
    fimg = load_image(funcfile)
    results = tsd.time_slice_diffs(fimg.get_data())
    # struct as record only to avoid deprecation warning
    tsd_results = sio.loadmat(pjoin(TEST_DATA_PATH, 'tsdiff_results.mat'),
                              struct_as_record=True, squeeze_me=True)
    assert_array_almost_equal(results['volume_means'], tsd_results['g'])
    assert_array_almost_equal(results['volume_mean_diff2'],
                              tsd_results['imgdiff'])
    assert_array_almost_equal(results['slice_mean_diff2'],
                              tsd_results['slicediff'])
    # next tests are from saved, reloaded volumes at 16 bit integer
    # precision, so are not exact, but very close, given that the mean
    # of this array is around 3200
    assert_array_almost_equal(results['diff2_mean_vol'],
                              tsd_results['diff2_mean_vol'],
                              decimal=1)
    assert_array_almost_equal(results['slice_diff2_max_vol'],
                              tsd_results['slice_diff2_max_vol'],
                              decimal=1)


def assert_arr_img_res(arr_res, img_res):
    for key in ('volume_mean_diff2',
                'slice_mean_diff2',
                'volume_means'):
        assert_array_equal(arr_res[key], img_res[key])
    for key in ('slice_diff2_max_vol', 'diff2_mean_vol'):
        assert_array_almost_equal(arr_res[key], img_res[key].get_data())


def test_tsd_image():
    # Test image version of time slice diff
    fimg = load_image(funcfile)
    data = fimg.get_data()
    tsda = tsd.time_slice_diffs
    tsdi = tsd.time_slice_diffs_image
    arr_results = tsda(data)
    # image routine insists on named slice axis, no default
    assert_raises(AxisError, tsdi, fimg)
    # Works when specifying slice axis as keyword argument
    img_results = tsdi(fimg, slice_axis='k')
    assert_arr_img_res(arr_results, img_results)
    ax_names = fimg.coordmap.function_domain.coord_names
    # Test against array version
    for time_ax in range(4):
        time_name = ax_names[time_ax]
        for slice_ax in range(4):
            slice_name = ax_names[slice_ax]
            if time_ax == slice_ax:
                assert_raises(ValueError, tsda, data, time_ax, slice_ax)
                assert_raises(ValueError, tsdi, fimg, time_ax, slice_ax)
                assert_raises(ValueError, tsdi, fimg, time_name, slice_ax)
                assert_raises(ValueError, tsdi, fimg, time_ax, slice_name)
                assert_raises(ValueError, tsdi, fimg, time_name, slice_name)
                continue
            arr_res = tsda(data, time_ax, slice_ax)
            assert_arr_img_res(arr_res, tsdi(fimg, time_ax, slice_ax))
            assert_arr_img_res(arr_res, tsdi(fimg, time_name, slice_ax))
            assert_arr_img_res(arr_res, tsdi(fimg, time_ax, slice_name))
            img_results = tsdi(fimg, time_name, slice_name)
            assert_arr_img_res(arr_res, img_results)
            exp_ax_names = tuple(n for n in ax_names if n != time_name)
            for key in ('slice_diff2_max_vol', 'diff2_mean_vol'):
                img = img_results[key]
                assert_equal(img.coordmap.function_domain.coord_names,
                             exp_ax_names)
    # Test defaults on rolled image
    fimg_rolled = rollimg(fimg, 't')
    # Still don't have a slice axis specified
    assert_raises(AxisError, tsdi, fimg_rolled)
    # Test default time axis
    assert_arr_img_res(arr_results, tsdi(fimg_rolled, slice_axis='k'))
    # Test axis named slice overrides default guess
    time_ax = -1
    for sa_no, sa_name in ((0, 'i'), (1, 'j'), (2, 'k')):
        fimg_renamed = fimg.renamed_axes(**{sa_name: 'slice'})
        arr_res = tsda(data, time_ax, sa_no)
        assert_arr_img_res(arr_res, tsdi(fimg_renamed, time_ax))
